# This example is for a dashboard user to modify data in a PostgreSQL DB table using 
# a modern grid configured with transaction functions. The workflow will rely on this 
# script to make actual modifications in the table. The table named USERS has 4 
# columns USER_ID(Int), USER_LOGIN(Varchar), HIRED_DATE(Date) and EMAIL(VarChar) 
# with USER_ID as the primary. Data changes for HIRED_DATE (Date) and EMAIL(VarChar) 
# are optional.

# In this template we use the library 'psycopg2' (installed in runtime as
# 'psycopg2-binary') to connect with the PostgreSQL database, submit atomic requests 
# and execute operations. For using other types of database, make sure to use
# their methods correctly. 

# Python libraries to handle SQL connections are not installed on the environment by
# default so it is necessary to choose runtime which have such library installed in the
# settings of the script. Additionally it is necessary to specify correct `Network
# Access` to allow script to communicate with some other IP then IP of the environment.
import psycopg2
import datetime


# helper method to handle not adding None values to the query
def append_not_none_value(array, value):
    if value is not None:
        array.append(value)
        return True
    return False

# Helper method to return datetime object in correct format for SQL operations
def parseDateTimeValue(dateValue):
    # somehow it will be converted to this datetime format
    #   "%Y-%m-%d %H:%M:%S.%f" -> 2023-04-01 00:00:00.000, 
    date_formats = [
        "%Y-%m-%d %H:%M:%S.%f",
        "%m/%d/%Y",
        "%m/%d/%Y %H:%M:%S",
        "%Y-%d-%m %H:%M:%S",
        "%m/%d/%Y %H:%M:%S.%f",
        "%m-%d-%Y"
    ]
    parsed_datetime = None
    for date_format in date_formats:
        try:
            parsed_datetime = datetime.datetime.strptime(dateValue, date_format)
            break
        except:
            pass

    if (parsed_datetime == None):
        raise Exception("Wrong Type of the datatime input value convertion!!!!")
    else:
        return parsed_datetime


try:
    # create connection to database. If you want to use other way then providing
    # user and password please check documentation of this method 
    connection = None
    connection = psycopg2.connect(
        host=$host,
        port=$port,
        database=$database,
        user=$user_name,
        password=$user_password
    )

    # Prepare queries for database operations. Due to the fact that for some rows
    # we might want to pass `None` value we need some additional logic. This logic
    # require to create query during the time of reading passed properties.

    # Parameterized query for insert new row
    # We assume that we must pass `USER_ID` and `USER_LOGIN` when inserting
    # and the other properties are optional.
    tmp_sql_insert_query = """
        INSERT INTO USERS (
            USER_ID,
            USER_LOGIN
    """

    # Parameterized query for edit the existing row
    # We assume that when updating we must pass only `USER_LOGIN` and the other
    # properties are optional. Update is done based on `USER_ID` and `WHERE` clause
    # is added in the end of preparing query for each row
    tmp_sql_update_query = """
        UPDATE USERS SET USER_LOGIN = %s
    """

    # Parameterized query for select the existing row to delete
    sql_delete_query = """DELETE FROM USERS WHERE USER_ID = %s"""

    # create cursor to execute operations on database
    cursor = connection.cursor()
    
    # deal with the every row data via prepared statements
    # we assume that the logic of passing data from dashboard works correctly and each
    # transactional column is either the array of length which is equal to number
    # of rows passed from dashboard or `None`. Values for `txn_actions` are calculated
    # by dashboard logic and they are based on the action which was created on
    # particular row.
    for index in range(len($txn_actions)):
        if $txn_actions[index] == "insert":
            # when value is optional and not provided, then don't add it to insert query
            optional_values_array = []
            optional_values_in_insert_query = ''
            # this value depends on number of the properties in initial INSERT query
            # it is equal to the number of properties there
            i = 2
            # when array with values for each property is not `None` and the value
            # for particular row is not `None` then we add it
            if $hired_date and $hired_date[index]:
                optional_values_array.append(parseDateTimeValue($hired_date[index]))
                optional_values_in_insert_query += f",HIRED_DATE"
                i = i + 1
            if $email and append_not_none_value(optional_values_array, $email[index]):
                optional_values_in_insert_query += f",EMAIL"
                i = i + 1

            # create list of parameters
            query_parameters = ",".join(list(map(lambda x: "%s", list(range(i)))))
            optional_values_in_insert_query += f") VALUES ({query_parameters})"

            sql_insert_tuple = (
                int($user_id[index]),
                $user_login[index],
            ) + tuple(optional_values_array)
            # final INSERT query is created based on initial INSERT query and the part
            # created for particular row based on properties passed for this row
            sql_insert_query = tmp_sql_insert_query + optional_values_in_insert_query

            # execute insert query
            cursor.execute(sql_insert_query, sql_insert_tuple)

        elif $txn_actions[index] == "update":
            optional_values_array = []
            optional_values_in_update_query = ''
            # when array with values for each property is not `None` and the value
            # for particular row is not `None` then we add it 
            if $hired_date and $hired_date[index]:
                optional_values_array.append(parseDateTimeValue($hired_date[index]))
                optional_values_in_update_query += f",HIRED_DATE = %s"
            if $email and append_not_none_value(optional_values_array, $email[index]):
                optional_values_in_update_query += f",EMAIL = %s"

            # final UPDATE query is created based on initial UPDATE query and the part
            # created for particular row based on properties passed for this row
            # after adding optional properties we need to add `WHERE` clause
            # which in this case is based on `USER_ID`
            optional_values_array.append(int($user_id[index]))
            optional_values_in_update_query += f" WHERE USER_ID = %s"

            sql_update_tuple = (
                $user_login[index],
            ) + tuple(optional_values_array)
            sql_update_query = tmp_sql_update_query + optional_values_in_update_query
            # execute update query
            cursor.execute(sql_update_query, sql_update_tuple)

        elif $txn_actions[index] == "select":
            # case with DELETE query is the easiest one because we are deleting
            # just based on `USER_ID`
            cursor.execute(sql_delete_query, tuple([int($user_id[index])]))

    # commit all changes when there weren't any errors
    connection.commit()

# error handling
except psycopg2.Error as e:
    # in case of any error we have to rollback all of the changes which
    # has been made on database
    if connection:
        connection.rollback()
    print("Error:", e)

finally:
    # close the database connection
    if connection:
        connection.close()
        


import logging
from dataclasses import dataclass
from enum import auto

from requests import Response

from mstrio import config
from mstrio.api import monitors
from mstrio.connection import Connection
from mstrio.utils.cache import Cache, ContentCacheMixin
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import Dictable, camel_to_snake
from mstrio.utils.version_helper import class_version_handler

logger = logging.getLogger(__name__)


class ContentCacheFormat(AutoName):
    EXCEL = auto()
    PDF = auto()
    CSV = auto()
    HTML = auto()
    XML = auto()
    PLAIN_TEXT = auto()
    BINARY_DEFINITION = auto()
    BINARY_DATA = auto()
    HTML5 = auto()
    FLASH = auto()
    TRANSACTION = auto()


@class_version_handler('11.3.0700')
class ContentCache(Cache, ContentCacheMixin):
    """
    Manage content cache.

    _CACHE_TYPE is a variable used by ContentCache class for cache filtering
    purposes.
    """

    _CACHE_TYPE = None

    def __init__(
        self,
        connection: 'Connection',
        id: str,
        content_cache_dict: dict | None = None,
    ) -> None:
        """Initialize the ContentCache object. If content_cache_dict is provided
        no I-Server request will be sent.

        Args:
            connection (Connection): MicroStrategy connection object returned by
                `connection.Connection()`.
            id (string): content cache id
            content_cache_dict (dict, optional): dictionary with properties of
                content cache object. If it is provided then cache dictionary
                will not be retrieved from I-Server.
        """
        super().__init__(connection, id)

        self._nodes = ContentCacheMixin.fetch_nodes(self._connection, self._project_id)
        if not content_cache_dict:
            self.fetch()
        else:
            self._init_variables(**content_cache_dict)

    def _init_variables(self, **kwargs) -> None:
        kwargs = camel_to_snake(kwargs)
        super()._init_variables(**kwargs)
        self._status = (
            ContentCacheStatus.from_dict(status)
            if (status := kwargs.get('status'))
            else None
        )
        self._format = (
            ContentCacheFormat(format_) if (format_ := kwargs.get('format')) else None
        )
        self._combined_id = (
            combined_id if (combined_id := kwargs.get('combined_id')) else None
        )
        self._chapter = kwargs.get('chapter')
        self._report_caches_used = kwargs.get('report_caches_used')
        self._host = kwargs.get('host')
        self._warehouse_tables_used = kwargs.get('warehouse_tables_used')
        self._last_load_time = kwargs.get('last_load_time')
        self._security_filter_id = kwargs.get('security_filter_id')
        self._prompt_answer = kwargs.get('prompt_answer')
        self._result_cache_type = kwargs.get('result_cache_type')
        self._expiration_time = kwargs.get('expiration_time')
        self._database_connection_id = kwargs.get('database_connection_id')
        self._database_login_id = kwargs.get('database_login_id')
        self._cache_location = kwargs.get('cache_location')
        self._location = kwargs.get('location')
        self._data_language = kwargs.get('data_language')
        self._metadata_language = kwargs.get('metadata_language')
        self._group_by = kwargs.get('group_by')
        self._xml_content = kwargs.get('xml_content', [])

    def __repr__(self) -> str:
        return f"ContentCache(id='{self.id}', source={self.source})"

    def fetch(self) -> None:
        """Fetches the cache from the server, refreshing the variables to match
        those currently stored on the server."""
        result = ContentCache.list_caches(
            connection=self._connection,
            project_id=self._project_id,
            nodes=self._nodes,
            id=self.id,
            to_dictionary=True,
        )
        if result:
            self._init_variables(**result[0])

    def __alter_status(
        self,
        op: str,
        value: bool | None = None,
        status: str | None = None,
        nodes: list[str] | None = None,
    ) -> Response | None:
        """Engine for altering ContentCache status

        Args:
            op (str): Replace or Remove operation to be performed
            value (bool, optional): Value used by operation [new value used to
                replace existing value]
            status (str, optional): Status on which the value should be changed
            nodes (list, optional): list of node names for the specified project

        Returns:
            Response object
        """
        logger_message = {
            'replace/loaded/True': 'load',
            'replace/loaded/False': 'unload',
            'remove/None/None': 'delete',
        }.get(f'{op}/{status}/{value}')
        if not self.combined_id and config.verbose:
            logger.info(
                f"Could not perform '{logger_message}' operation on deleted cache "
                f"with ID '{self.id}'."
            )
            if logger_message == 'load':
                raise ValueError(
                    f"Could not perform 'load' operation on deleted cache "
                    f"with ID '{self.id}'."
                )
            return
        if not nodes:
            nodes = ContentCacheMixin.fetch_nodes(
                self._connection, self._connection.project_id
            )
        body = {
            'operationList': [
                {
                    'op': op,
                    'path': (
                        f'/contentCaches/{self.combined_id}/status/{status}'
                        if status
                        else f'/contentCaches/{self.combined_id}'
                    ),
                    'value': value,
                }
            ]
        }
        return monitors.update_contents_caches(self._connection, nodes, body)

    def load(self) -> bool:
        """Load content cache."""
        nodes = self._nodes
        result = self.__alter_status(
            op='replace', value=True, status='loaded', nodes=nodes
        )

        if config.verbose and result:
            logger.info(f'Successfully loaded content cache with id: {self.id}')
        if result:
            self.fetch()
            return True

    def unload(self) -> bool:
        """Unload content cache."""
        nodes = self._nodes
        result = self.__alter_status(
            op='replace', value=False, status='loaded', nodes=nodes
        )
        if config.verbose and result:
            logger.info(f'Successfully unloaded content cache with id: {self.id}')
        if result:
            self.fetch()
            return True
        return False

    def delete(self, force: bool | None = None) -> Response | None:
        """Delete content cache.

        Args:
            force (bool, optional): If True, then no additional prompt will be
                shown before deleting object.

        Returns:
            Response object."""
        if not force:
            user_input = (
                input(
                    f"Are you sure you want to delete content cache "
                    f"with ID: '{self.id}'? [Y/N]: "
                )
                or 'N'
            )
        if force or user_input == 'Y':
            nodes = self._nodes
            result = self.__alter_status(op='remove', nodes=nodes)
            if config.verbose and result:
                logger.info(f"Successfully deleted content cache with ID: '{self.id}'.")
            return result

    @classmethod
    def from_dict(
        cls, connection: 'Connection', caches: list[dict]
    ) -> list['ContentCache']:
        """Creates Caches from a provided dictionary.

        Args:
            cls (object): Class type for objects to be created
            connection (Connection): MicroStrategy connection object returned by
                `connection.Connection()`
            caches (list[dict]): list of dictionaries the Caches will be created
                from

        Returns:
            List of Caches created from the provided dictionaries."""
        return [
            ContentCache(connection, cache_dict['id'], cache_dict)
            for cache_dict in caches
        ]

    def list_properties(self) -> dict:
        """List properties for content cache."""
        return {
            **super().list_properties(),
            'status': self.status,
            'format': self.format,
            'chapter': self.chapter,
            'report_cache_used': self.report_caches_used,
            'host': self.host,
            'warehouse_tables_used': self.warehouse_tables_used,
            'last_load_time': self.last_load_time,
            'security_filter_id': self.security_filter_id,
            'prompt_answer': self.prompt_answer,
            'result_cashe_type': self.result_cache_type,
            'expiration_time': self.expiration_time,
            'database_connection_id': self.database_connection_id,
            'database_login_id': self.database_login_id,
            'cache_location': self.cache_location,
            'location': self.location,
            'data_language': self.data_language,
            'metadata_language': self.metadata_language,
            'group_by': self.group_by,
            'xml_content': self.xml_content,
        }

    @property
    def status(self):
        return self._status

    @property
    def format(self):
        return self._format

    @property
    def combined_id(self):
        return self._combined_id

    @property
    def chapter(self):
        return self._chapter

    @property
    def report_caches_used(self):
        return self._report_caches_used

    @property
    def host(self):
        return self._host

    @property
    def warehouse_tables_used(self):
        return self._warehouse_tables_used

    @property
    def last_load_time(self):
        return self._last_load_time

    @property
    def security_filter_id(self):
        return self._security_filter_id

    @property
    def prompt_answer(self):
        return self._prompt_answer

    @property
    def result_cache_type(self):
        return self._result_cache_type

    @property
    def expiration_time(self):
        return self._expiration_time

    @property
    def database_connection_id(self):
        return self._database_connection_id

    @property
    def database_login_id(self):
        return self._database_login_id

    @property
    def cache_location(self):
        return self._cache_location

    @property
    def location(self):
        return self._location

    @property
    def data_language(self):
        return self._data_language

    @property
    def metadata_language(self):
        return self._metadata_language

    @property
    def group_by(self):
        return self._group_by

    @property
    def xml_content(self):
        return self._xml_content


@dataclass
class ContentCacheStatus(Dictable):
    ready: bool
    processing: bool
    invalid: bool
    expired: bool
    loaded: bool
    filed: bool
    dirty: bool



import logging
from enum import auto
from time import sleep
from typing import Optional

from mstrio import config
from mstrio.api import migration
from mstrio.connection import Connection
from mstrio.types import ObjectTypes
from mstrio.users_and_groups.user import User
from mstrio.utils import helper
from mstrio.utils.entity import DeleteMixin, EntityBase
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import Dictable, exception_handler
from mstrio.utils.wip import WipLevels, module_wip

module_wip(globals(), level=WipLevels.PREVIEW)

logger = logging.getLogger(__name__)

TIMEOUT = 60
TIMEOUT_INCREMENT = 0.25


class PackageSettings(Dictable):
    """Object representation of package settings details.

    Attributes:
        default_action(DefaultAction): default action for Package
        update_schema(Optional[UpdateSchema]): update_schema for Package
        acl_on_replacing_objects(Optional[AclOnReplacingObjects]):
            ACL setting on replacing objects
        acl_on_new_objects(Optional[AclOnNewObjects]):
            ACL setting on new objects
    """

    class UpdateSchema(AutoName):
        """They allow you to configure the package to automatically perform
        certain schema update functions. These options can be useful if you make
        any changes to schema objects. Use the recalculate table keys and fact
        entry levels if you changed the key structure of a table or if you
        changed the level at which a fact is stored.Use the recalculate table
        logical sizes to override any modifications that you have made to
        logical table sizes. (Logical table sizes affect how the MicroStrategy
        SQL Engine determines which tables to use in a query.)
        """

        RECAL_TABLE_LOGICAL_SIZE = auto()
        RECAL_TABLE_KEYS_FACT_ENTRY_LEVEL = auto()

    class AclOnReplacingObjects(AutoName):
        """If you resolve a conflict with the "Replace" action, and the access
        control lists (ACL) of the objects are different between the two
        projects, you can choose whether to keep the existing ACL in the
        destination project or replace it with the ACL from the source project.
        Note: This is not supported for project security packages.
        """

        USE_EXISTING = auto()
        REPLACE = auto()

    class AclOnNewObjects(AutoName):
        """If you add a new object to the destination project with the
        "Create New" or "Keep Both action", you can choose to have the object
        inherit its ACL from the destination folder instead of keeping its own
        ACL. This is helpful when copying an object into a user's profile
        folder, so that the user can have full control over the object.
        """

        KEEP_ACL_AS_SOURCE_OBJECT = auto()
        INHERIT_ACL_AS_DEST_FOLDER = auto()

    class DefaultAction(AutoName):
        """The default action used for objects which don't have actions
        explicitly, for example the dependents objects.
        USE_EXISTING: No change is made to the destination object.
        The source object is not copied.
        REPLACE: The destination object is replaced with the source object.
        Note the following:
            • If the conflict type is Exists identically except for path, or
            Exists identically except for Distribution Services objects,
            the  destination object is updated to reflect the path or
            Distribution Services addresses and contacts of the source object.
            • Replace moves the object into the same parent folder as source
            object. If the parent path is the same between source and
            destination but the grandparent path is different, Replace may
            appear to do nothing because Replace puts the object into the same
            parent path.
            • Non-empty folders in the destination location will never have
            the same version ID and modification time as the source, because
            the folder is copied first and the objects are added to it, thus
            changing version ID and modification times during the copy process.
        KEEP_BOTH: No change is made to the destination object. The source
        object is duplicated if destination object doesn't exist. But if the
        destination object exists with the same id and same version, this source
        object is ignored.If the destination object exists with the same id
        and different version, this source object is saved as new object.
        USE_NEWER: If the source object's modification time is more recent than
        the destination object's, the Replace action is used.Otherwise, the Use
        existing action is used.
        USE_OLDER: If the source object's modification time is more recent than
        the destination object's, the Use existing action is used. Otherwise,
        the Replace action is used.
        FORCE_REPLACE: Replace the object in the destination project with the
        version of the object in the update package, even if both versions of
        the object have the same Version ID.
        DELETE: Delete the object from the destination project. The version
        of the object in the update package is not imported into the destination
        project.Warning: If the object in the destination has any used-by
        dependencies when you import the update package, the import will fail.
        FORCE_KEEP_BOTH: No change is made to the destination object. The source
        object is always saved as a new object.
        """

        USE_EXISTING = auto()
        REPLACE = auto()
        KEEP_BOTH = auto()
        USE_NEWER = auto()
        USE_OLDER = auto()
        FORCE_REPLACE = auto()
        DELETE = auto()
        FORCE_KEEP_BOTH = auto()

    _FROM_DICT_MAP = {
        'default_action': DefaultAction,
        'update_schema': [UpdateSchema],
        'acl_on_replacing_objects': AclOnReplacingObjects,
        'acl_on_new_objects': [AclOnNewObjects],
    }

    def __init__(
        self,
        default_action: DefaultAction = DefaultAction.USE_EXISTING,
        update_schema: UpdateSchema | None = None,
        acl_on_replacing_objects: AclOnReplacingObjects = (
            AclOnReplacingObjects.USE_EXISTING
        ),
        acl_on_new_objects: AclOnNewObjects = AclOnNewObjects.KEEP_ACL_AS_SOURCE_OBJECT,
    ):
        self.default_action = default_action
        self.update_schema = (
            update_schema
            if isinstance(update_schema, list) or not update_schema
            else [update_schema]
        )
        self.acl_on_replacing_objects = acl_on_replacing_objects
        self.acl_on_new_objects = (
            [acl_on_new_objects]
            if isinstance(acl_on_new_objects, PackageSettings.AclOnNewObjects)
            else acl_on_new_objects
        )


class PackageContentInfo(Dictable):
    """Object representation of package content information

    Attributes:
        id(str): object ID
        action(Action): The action to resolve the conflict.
        name(Optional[str]): object name
        version(Optional[str]): object version
        type(Optional[ObjectTypes]): object type
        owner(Optional[Owner]): owner of object
        date_created(Optional[str]): object creation date
        date_modified(Optional[str]): object modification date
        include_dependents(Optional[bool]): whether include the dependents
            for this object or not
        explicit_included(Optional[bool]): whether explicitly included or not
        level(Optional[Level]): level of object
    """

    class Action(AutoName):
        """The default action used for objects which don't have actions
        explicitly, for example the dependents objects.
        USE_EXISTING: No change is made to the destination object.
        The source object is not copied.
        REPLACE: The destination object is replaced with the source object.
        Note the following:
            • If the conflict type is Exists identically except for path, or
            Exists identically except for Distribution Services objects,
            the  destination object is updated to reflect the path or
            Distribution Services addresses and contacts of the source object.
            • Replace moves the object into the same parent folder as source
            object. If the parent path is the same between source and
            destination but the grandparent path is different, Replace may
            appear to do nothing because Replace puts the object into the same
            parent path.
            • Non-empty folders in the destination location will never have
            the same version ID and modification time as the source, because
            the folder is copied first and the objects are added to it, thus
            changing version ID and modification times during the copy process.
        KEEP_BOTH: No change is made to the destination object. The source
        object is duplicated if destination object doesn't exist. But if the
        destination object exists with the same id and same version, this source
        object is ignored.If the destination object exists with the same id
        and different version, this source object is saved as new object.
        USE_NEWER: If the source object's modification time is more recent than
        the destination object's, the Replace action is used.Otherwise, the Use
        existing action is used.
        USE_OLDER: If the source object's modification time is more recent than
        the destination object's, the Use existing action is used. Otherwise,
        the Replace action is used.
        FORCE_REPLACE: Replace the object in the destination project with the
        version of the object in the update package, even if both versions of
        the object have the same Version ID.
        DELETE: Delete the object from the destination project. The version
        of the object in the update package is not imported into the destination
        project.Warning: If the object in the destination has any used-by
        dependencies when you import the update package, the import will fail.
        FORCE_KEEP_BOTH: No change is made to the destination object. The source
        object is always saved as a new object.
        """

        USE_EXISTING = auto()
        REPLACE = auto()
        KEEP_BOTH = auto()
        USE_NEWER = auto()
        USE_OLDER = auto()
        FORCE_REPLACE = auto()
        DELETE = auto()
        FORCE_KEEP_BOTH = auto()

    class Level(AutoName):
        PROJECT_OBJECT = auto()
        CONFIGURATION_OBJECT = auto()
        UNKNOWN = auto()

    class Owner(User):
        """
        This class overrides `from_dict()` and `__init__()` methods as
        PackageContentInfo inherits from `Dictable` and not `EntityBase`.
        Therefore, despite `Owner` being a `User`, `connection` param here is
        optional. If `connection` is specified, `User` constructor will be used
        """

        @classmethod
        def from_dict(
            cls, source: dict[str, any], connection: Optional["Connection"] = None
        ):
            return super().from_dict(source, connection)

        def __init__(self, id: str, connection: Optional["Connection"] = None):
            if connection:
                super().__init__(id, connection)
            else:
                self._id = id

        def _init_variables(self, **kwargs) -> None:
            self._connection = kwargs.get("connection")
            self._id = kwargs.get("id")

    _FROM_DICT_MAP = {
        'type': ObjectTypes,
        'action': Action,
        'level': Level,
        'owner': Owner,
    }

    def __init__(
        self,
        id: str,
        action: Action | str = Action.USE_EXISTING,
        name: str | None = None,
        version: str | None = None,
        type: ObjectTypes | None = None,
        owner: Owner | None = None,
        date_created: str | None = None,
        date_modified: str | None = None,
        include_dependents: bool | None = None,
        explicit_included: bool | None = None,
        level: Level | str | None = None,
    ):
        self.id = id
        self.name = name
        self.version = version
        self.type = type
        self.owner = owner
        self.date_created = date_created
        self.date_modified = date_modified
        self.include_dependents = include_dependents
        self.explicit_included = explicit_included
        try:
            self.level = (
                PackageContentInfo.Level(level) if isinstance(level, str) else level
            )
            self.action = (
                PackageContentInfo.Action(action) if isinstance(action, str) else action
            )
        except ValueError:
            exception_handler(msg="Wrong enum value", exception_type=ValueError)


class PackageConfig(Dictable):
    """Package Update Data Transfer Object

    Attributes:
        type(PackageUpdateType): type of package update
        settings(PackageSettings): settings details of package
        content(PackageContentInfo, List[PackageContentInfo]): content details
            of package
    """

    class PackageUpdateType(AutoName):
        """Package update type:
        PROJECT: For user’s input, only accept non configuration object. But
            the actual package contains all kinds of objects, including
            configuration objects.
        CONFIGURATION: Only contains configuration objects. 12 types in total,
            including “Database connection“, “Transmitter“, “Database Instance“,
            “Database login“, “DBMS“, “Device“, “Event“, ”Language”, “Schedule“,
            ”Security Role”,”User”, “User group“.
        PROJECT_SECURITY: Only contains user objects.
        """

        PROJECT = auto()
        PROJECT_SECURITY = auto()
        CONFIGURATION = auto()

    _FROM_DICT_MAP = {
        'type': PackageUpdateType,
        'settings': PackageSettings,
        'content': [PackageContentInfo],
    }

    def __init__(
        self,
        type: PackageUpdateType,
        settings: PackageSettings,
        content: list[PackageContentInfo] | PackageContentInfo,
    ):
        self.type = type
        self.settings = settings
        self.content = [content] if not isinstance(content, list) else content


class Package(EntityBase, DeleteMixin):
    """Object representation of MicroStrategy Package object.

    Attributes:
        connection(Connection): a MicroStrategy connection object
        id(str): package ID
        status(str): status of a package
        settings(PackageSettings): settings details of package
        content(PackageContentInfo): content details of package
    """

    _API_GETTERS = {
        ("id", "status", "settings", "content"): migration.get_package_holder
    }
    _API_DELETE = staticmethod(migration.delete_package_holder)

    _FROM_DICT_MAP = {
        'settings': PackageSettings.from_dict,
        'content': [PackageContentInfo.from_dict],
    }

    def __init__(self, connection: Connection, id: str) -> None:
        """Initialize PackageImport object.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`
            id: ID of package import process
        """
        if id is None:
            exception_handler(
                "Please provide actual value for id argument, other than None.",
                exception_type=ValueError,
            )

        super().__init__(connection, id)

    def _init_variables(self, **kwargs) -> None:
        self.settings = None
        self.content = None

        super()._init_variables(**kwargs)
        self.status = kwargs.get("status")
        self.project_id = kwargs.get("project_id")

    @classmethod
    def create(cls, connection: Connection, progress_bar: bool = False):
        """Create a Package object.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`
            progress_bar: boolean value that decides whether a progress bar will
                be displayed. defaults to False

        Returns:
            A Package object."""
        response = migration.create_package_holder(connection).json()
        if config.verbose and not progress_bar:
            logger.info(f"Created package with ID: '{response.get('id')}'")
        return cls.from_dict(source=response, connection=connection)

    def update_config(self, package_config: PackageConfig):
        """Updates the config for the Package.

        Args:
            package_config: new configuration to be updated for the Package

        Returns:
            A boolean value of whether the update was successful."""
        body = package_config.to_dict()
        body = helper.delete_none_values(body, recursion=True)
        body = helper.snake_to_camel(body)
        migration.update_package_holder(self.connection, body, self.id).json()

        total_time = 0
        # Wait until update ready
        while True:
            response = migration.get_package_holder(
                self.connection, self.id, show_content=False
            ).json()
            if response.get('status') == 'ready':
                break
            sleep(TIMEOUT_INCREMENT)
            total_time += TIMEOUT_INCREMENT

            if total_time > TIMEOUT:
                logger.warning('Time out on updating package')
                return False

        self.fetch()
        return True

    def upload_package_binary(self, package_binary: bytes, progress_bar: bool = False):
        """Uploads a binary of a package.

        Args:
            package_binary: binary of the package to be uploaded
            progress_bar: boolean value that decides whether a progress bar will
                be displayed. defaults to False"""
        response = migration.upload_package(
            self.connection, self.id, package_binary
        ).json()
        self.status = response.get('status')
        if config.verbose and not progress_bar:
            logger.info(
                f"Uploaded package binary to package holder with ID: "
                f"'{response.get('id')}'"
            )

    def download_package_binary(self, progress_bar: bool = False) -> bytes:
        """Downloads a binary of a package.

        Args:
            progress_bar: boolean value that decides whether a progress bar will
                be displayed. defaults to False

        Returns:
            Contents of the downloaded binary of the package."""
        response = migration.download_package(self.connection, self.id)
        if config.verbose and not progress_bar:
            logger.info(f"Downloaded binary of a package with ID: '{self.id}'")
        return response.content


class PackageImport(EntityBase, DeleteMixin):
    """Object representation of MicroStrategy PackageImportProcess object.

    Attributes:
        connection(Connection): A MicroStrategy connection object
        id(str): PackageImport ID
        status(str): status of an import
        undo_package_created(bool): if the undo package have been created
        progress(int): progress of package import process
    """

    _API_GETTERS = {
        ("id", "status", "undo_package_created", "progress"): migration.get_import
    }
    _API_DELETE = staticmethod(migration.delete_import)
    _progress_bar = True

    def __init__(self, connection: Connection, id: str) -> None:
        """Initialize PackageImport object.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            id: ID of package import process
        """
        if id is None:
            exception_handler(
                "Please provide actual value for id argument, other than None.",
                exception_type=ValueError,
            )

        super().__init__(connection, id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.undo_package_created = kwargs.get("undo_package_created")
        self.status = kwargs.get("status")
        self.progress = kwargs.get("progress")

    @classmethod
    def create(
        cls,
        connection: Connection,
        package_id: str,
        generate_undo: bool,
        progress_bar=False,
    ):
        """Create a package import process.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            package_id: ID of the package that is to be imported
            generate_undo: boolean value that specifies whether an undo package
                has to also be created
            progress_bar: boolean value that decides whether a progress bar will
                be displayed. defaults to False

        Returns:
            A PackageImport object."""
        response = migration.create_import(
            connection, package_id, generate_undo=generate_undo
        ).json()

        total_time = 0
        # Wait until update ready
        while True:
            response = migration.get_import(connection, response.get('id')).json()
            if response.get('status') == 'imported':
                break
            sleep(TIMEOUT_INCREMENT)
            total_time += TIMEOUT_INCREMENT

            if total_time > TIMEOUT:
                logger.warning('Time out on creating import')
                return False

        if config.verbose and not progress_bar:
            logger.info(
                f"Created package import process with ID: '{response.get('id')}'"
            )
        return cls.from_dict(source=response, connection=connection)

    def download_undo_binary(self, progress_bar=False):
        """Download undo package binary for this import process.

        Args:
        progress_bar: boolean value that decides whether a progress bar will
            be displayed. defaults to False

        Returns:
            Contents of the Response object containing all of the information
        returned by the server."""
        total_time = 0
        # Wait until update ready
        while True:
            response = migration.get_import(self.connection, self.id).json()
            if response.get('undoPackageCreated') is True:
                break
            sleep(TIMEOUT_INCREMENT)
            total_time += TIMEOUT_INCREMENT

            if total_time > TIMEOUT:
                logger.warning('Time out on downloading undo binary')
                return False

        response = migration.create_undo(self.connection, self.id)
        if config.verbose and not progress_bar:
            logger.info(
                f"Downloaded undo package binary for import process with ID: "
                f"'{self.id}'"
            )
        return response.content



import logging
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.api import datasources
from mstrio.users_and_groups.user import User
from mstrio.utils import helper
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, ObjectTypes
from mstrio.utils.helper import get_args_from_func, get_default_args_from_func
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.translation_mixin import TranslationMixin
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.connection import Connection

logger = logging.getLogger(__name__)


@method_version_handler('11.2.0500')
def list_datasource_logins(
    connection: "Connection", to_dictionary: bool = False, limit: int = None, **filters
) -> list["DatasourceLogin"] | list[dict]:
    """Get list of DatasourceLogin objects or dicts. Optionally filter the
    logins by specifying filters.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        to_dictionary: If True returns dict, by default (False) returns
            User objects.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'acg', 'username']

    Examples:
        >>> list_db_logins(connection, name='db_login_name')
    """
    return DatasourceLogin._list_datasource_logins(
        connection=connection,
        to_dictionary=to_dictionary,
        limit=limit,
        **filters,
    )


@class_version_handler('11.2.0500')
class DatasourceLogin(Entity, CopyMixin, DeleteMixin, TranslationMixin):
    """A user login configuration object to access a particular datasource. Also
    formerly known as database login.

    Attributes:
        name: name of the database login
        id: database login ID
        description: Description of the database login
        username: database user to be used by the database login
        version: Object version ID
        ancestors: List of ancestor folders
        type: Object type. Enum
        subtype: Object subtype
        ext_type: Object extended type.
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        owner: User object that is the owner
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
    """

    _OBJECT_TYPE = ObjectTypes.DBLOGIN
    _FROM_DICT_MAP = {**Entity._FROM_DICT_MAP, 'owner': User.from_dict}
    _API_GETTERS = {
        (
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acl',
            'comments',
        ): objects_processors.get_info,
        (
            'id',
            'name',
            'description',
            'username',
            'date_created',
            'date_modified',
            'acg',
        ): datasources.get_datasource_login,
    }
    _API_PATCH: dict = {
        ('abbreviation', 'comments'): (
            objects_processors.update,
            'partial_put',
        ),
        ("name", "username", "description", "password"): (
            datasources.update_datasource_login,
            "patch",
        ),
    }
    _PATCH_PATH_TYPES = {
        "name": str,
        "username": str,
        "description": str,
        "password": str,
    }

    def __init__(self, connection: "Connection", name: str = None, id: str = None):
        """Initialize DatasourceLogin object."""

        if id is None and name is None:
            raise ValueError(
                "Please specify either 'id' or 'name' parameter in the constructor."
            )

        if id is None:
            objects_info = DatasourceLogin._list_datasource_logins(
                connection=connection, name=name, to_dictionary=True
            )
            if objects_info:
                object_info, object_info["connection"] = objects_info[0], connection
                self._init_variables(**object_info)
            else:
                raise ValueError(f"There is no Datasource Login: '{name}'")
        else:
            super().__init__(connection=connection, object_id=id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.username = kwargs.get("username")

    def alter(
        self,
        name: str = None,
        username: str = None,
        description: str = None,
        password: str = None,
        comments: str | None = None,
    ) -> None:
        """Alter the datasource login properties.

        Args:
            name: login object name
            username: username
            description: login object description
            password: database password to be used by the database login
            comments: long description of the login object
        """
        func = self.alter
        args = get_args_from_func(func)
        defaults = get_default_args_from_func(func)
        default_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    @classmethod
    def create(
        cls,
        connection: "Connection",
        name: str,
        username: str,
        password: str,
        description: str = None,
    ) -> "DatasourceLogin":
        """Create a new datasource login.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`
            name: login object name
            username: username
            description: login object description
            password: database password to be used by the database login
        Returns:
            DatasourceConnection object.
        """
        body = {
            "name": name,
            "description": description,
            "username": username,
            "password": password,
        }
        body = helper.delete_none_values(body, recursion=True)
        response = datasources.create_datasource_login(connection, body).json()
        if config.verbose:
            logger.info(
                f"Successfully created datasource login named: '"
                f"{response.get('name')}' "
                f"with ID: '{response.get('id')}'"
            )
        return cls.from_dict(source=response, connection=connection)

    @classmethod
    def _list_datasource_logins(
        cls,
        connection: "Connection",
        to_dictionary: bool = False,
        limit: int = None,
        **filters,
    ) -> list["DatasourceLogin"] | list[dict]:
        objects = helper.fetch_objects(
            connection=connection,
            api=datasources.get_datasource_logins,
            dict_unpack_value="logins",
            limit=limit,
            filters=filters,
        )
        if to_dictionary:
            return objects
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]



from enum import Enum, IntEnum


class CertifiedStatus(Enum):
    """Enumeration that represents what can be passed in the certified_status
    attribute of the IServer quick search command."""

    ALL = 'ALL'
    CERTIFIED_ONLY = 'CERTIFIED_ONLY'
    NOT_CERTIFIED_ONLY = 'NOT_CERTIFIED_ONLY'
    OFF = 'OFF'


class SearchPattern(IntEnum):
    """Enumeration constants used to specify searchType used to control BI
    Search. More details can be found in EnumDSSXMLSearchTypes in a browser."""

    CONTAINS_ANY_WORD = 0
    BEGIN_WITH = 1
    EXACTLY = 2
    BEGIN_WITH_PHRASE = 3
    CONTAINS = 4
    END_WITH = 5


class SearchDomain(IntEnum):
    """Enumeration constants used to specify the search domains. More details
    can be found in EnumDSSXMLSearchDomain in a browser."""

    LOCAL = 1
    PROJECT = 2
    REPOSITORY = 3
    CONFIGURATION = 4
    CONFIGURATION_AND_ALL_PROJECTS = 5


class SearchResultsFormat(Enum):
    """Enumeration constants used to specify the format to return
    from search functions."""

    LIST = 'LIST'
    TREE = 'TREE'



# flake8: noqa
from .content_cache import ContentCache
from .datasets import *

# isort: off
from .dashboard import Dashboard, list_dashboards, list_dashboards_across_projects

# isort: on
from .document import Document, list_documents, list_documents_across_projects
from .dossier import (
    ChapterPage,
    Dossier,
    PageSelector,
    PageVisualization,
    VisualizationSelector,
    list_dossiers,
    list_dossiers_across_projects,
)
from .library import Library
from .report import Report, list_reports



# This example is for a dashboard user to modify data in a Rally workspace using
# a modern grid configured with transaction functions. The workflow will rely on this
# script to make actual modifications in the defect. The properties available for
# modification in the defect are `Name`, `Description`, `Priority`, `Owner`,
# `Iteration`, `TargetDate`, `State` and `PlanEstimate`. Only the `Name` property is
# required. Property  `FormattedID` cannot be modified.

# To connect with Rally API we use the library 'pyral'. Python library to handle Rally
# connection is not installed on the environment by default so it is necessary to choose
# runtime which have such library installed in the settings of the script. Additionally
# it is necessary to specify correct 'Network Access' to allow script to communicate
# with some other IP then IP of the environment
from pyral import Rally, restapi
from datetime import datetime

# Helper method to return datetime object in correct format for Rally operations
def parse_date_time_value(dateValue):
    # somehow it will be converted to this datetime format
    #   "%Y-%m-%d %H:%M:%S.%f" -> 2023-04-01 00:00:00.000
    date_formats = [
        "%Y-%m-%d %H:%M:%S.%f",
        "%Y-%m-%d %H:%M:%S",
        "%Y-%m-%d",
        "%m/%d/%Y %H:%M:%S.%f",
        "%m/%d/%Y %H:%M:%S",
        "%m/%d/%Y",
    ]
    parsed_datetime = None
    for date_format in date_formats:
        try:
            parsed_datetime = datetime.strptime(dateValue, date_format)
            break
        except Exception:
            pass

    if parsed_datetime is None:
        raise Exception("Wrong Type of the datatime input value convertion!!!!")
    else:
        print(parsed_datetime)
        return parsed_datetime

try:
    # create connection to Rally. If you want to use other way then providing
    # api key please check documentation of this method 
    rally = Rally(
        server=$rally_server,
        apikey=$api_key,
        workspace=$workspace,
        project=$project,
    )

    # deal with the every row data via prepared statements
    # we assume that the logic of passing data from dashboard works correctly and each
    # transactional column is either the array of length which is equal to number
    # of rows passed from dashboard or `None`.
    for index in range(len($formatted_id)):
        # UPDATE RALLY DEFECT
        tmp_formatted_id = $formatted_id[index]
        # first it is necessary to find object which will be modified
        defect = response_de = rally.get(
            "Defect",
            fetch=True,
            query=f"FormattedID = {tmp_formatted_id}",
            projectScopeDown=True,
        ).next()
        update_data = {"ObjectID": defect.ObjectID}
        if $defect_name and $defect_name[index]:
            update_data["Name"] = $defect_name[index]
        if $description:  # Value for description can be None
            update_data["Description"] = $description[index]
        # Valid 'Priority' values are [ Low, Normal, High Attention,
        # Resolve Immediately, None ]
        if $priority:
            update_data["Priority"] = $priority[index]
        # Value for owner can be None or existing owner username
        if $owner:
            owner = None
            # inside defect owner has to be stored as a reference to the actual user
            # value passed from dashboard is used to find this reference
            if $owner[index]:
                owner_username = $owner[index]
                owner = rally.get(
                    "User", fetch=True, query=f"UserName = {owner_username}"
                ).next().ref
            update_data["Owner"] = owner
        # Value for iteration can be None or existing iteration name
        if $iteration:
            iteration = None
            if $iteration[index]:
                tmp_iteration_name = $iteration[index]
                iteration = rally.get(
                    "Iteration", fetch=True, query=f'Name = "{tmp_iteration_name}"'
                ).next().ref
            update_data["Iteration"] = iteration
        if $target_date:
            target_date = (
                None
                if $target_date[index] is None
                else parse_date_time_value($target_date[index]).isoformat()
            )
            update_data["TargetDate"] = target_date
        # Valid 'State' values are: [ Submitted, Closed, Deferred, Open ]
        if $state and $state[index]:
            update_data["State"] = $state[index]
        if $plan_estimate:  # Value for 'Plan Estimate' can be None
            update_data["PlanEstimate"] = $plan_estimate[index]

        rally.update("Defect", update_data)
        
except restapi.RallyRESTAPIError as error:
    raise Exception(f"Error has occurred: {error}")
# this error is raised from `rally.get` when the object was not found and we are trying
# to iterate with `next` method
except StopIteration as error:
    raise Exception(f"Error while iterating has ocurred: {error}")



import json
import logging
from datetime import datetime
from enum import Enum, IntFlag
from typing import TYPE_CHECKING

import pandas as pd
from pandas import DataFrame, read_csv
from requests.exceptions import HTTPError

from mstrio import config
from mstrio.access_and_security.privilege_mode import PrivilegeMode
from mstrio.access_and_security.security_role import SecurityRole
from mstrio.connection import Connection
from mstrio.server.language import Language
from mstrio.users_and_groups.user_connections import UserConnections
from mstrio.utils import helper, time_helper
from mstrio.utils.acl import TrusteeACLMixin
from mstrio.utils.entity import DeleteMixin, Entity, ObjectTypes
from mstrio.utils.helper import Dictable, filter_params_for_func
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.response_processors import users
from mstrio.utils.translation_mixin import TranslationMixin
from mstrio.utils.version_helper import method_version_handler

if TYPE_CHECKING:
    from mstrio.access_and_security.privilege import Privilege
    from mstrio.modeling.security_filter import SecurityFilter
    from mstrio.server.project import Project
    from mstrio.users_and_groups.user_group import UserGroup

logger = logging.getLogger(__name__)


def create_users_from_csv(connection: Connection, csv_file: str) -> list["User"]:
    """Create new user objects from csv file. Possible header values for the
    users are the same as in the `User.create()` method.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        csv_file: path to file containing at minimum 'username' and 'full_name'
            headers'.
    """
    return User._create_users_from_csv(connection=connection, csv_file=csv_file)


def list_users(
    connection: Connection,
    name_begins: str | None = None,
    abbreviation_begins: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list["User"] | list[dict]:
    """Get list of user objects or user dicts. Optionally filter the users by
    specifying 'name_begins', 'abbreviation_begins' or other filters.

    Wildcards available for name_begins and abbreviation_begins:
        ? - any character
        * - 0 or more of any characters
        e.g. name_begins = ?onny will return Sonny and Tonny

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        name_begins: characters that the user name must begin with.
        abbreviation_begins: characters that the abbreviation must begin with.
        to_dictionary: If True returns dict, by default (False) returns
            User objects.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'abbreviation',
            'description', 'type', 'subtype', 'date_created', 'date_modified',
            'version', 'acg', 'icon_path', 'owner', 'initials', 'enabled']

    Examples:
        >>> list_users(connection, name_begins='user', initials='UR')
    """
    return User._get_users(
        connection=connection,
        name_begins=name_begins,
        abbreviation_begins=abbreviation_begins,
        to_dictionary=to_dictionary,
        limit=limit,
        **filters,
    )


class User(Entity, DeleteMixin, TrusteeACLMixin, TranslationMixin):
    """Object representation of MicroStrategy User object.

    Attributes:
        connection: A MicroStrategy connection object
        id: User ID
        name: User name
        username: User username
        full_name: full name of the User
        initials: User initials, derived from user's last name or username
        abbreviation: User login name
        description: User description
        memberships: IDs and names of direct parent groups for user
        security_roles: security roles that the user is a member of
        addresses: addresses for the user
        privileges: user privileges per project
        trust_id: Unique user ID provided by trusted authentication provider
        enabled: Specifies if user is allowed to log in
        owner: owner ID and name
        ancestors: List of ancestor folders
        password_modifiable: If user password can be modified
        require_new_password: If user is required to change new password
        standard_auth: If standard authentication is allowed for user
        ldapdn: information about user's LDAP distinguished name
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        type: Object type
        subtype: Object subtype
        ext_type: Object extended type
        version: Version ID
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
        default_timezone:  Information about user default timezone
        language: Information about user language
    """

    _PATCH_PATH_TYPES = {
        "name": str,
        "username": str,
        "full_name": str,
        "description": str,
        "password": str,
        "enabled": bool,
        "password_modifiable": bool,
        "password_expiration_date": str,
        "standard_auth": bool,
        "require_new_password": bool,
        "ldapdn": str,
        "trust_id": str,
        "abbreviation": str,
        "owner": str,
    }
    _OBJECT_TYPE = ObjectTypes.USER
    _API_GETTERS = {
        (
            'id',
            'name',
            'type',
            'subtype',
            'ext_type',
            'abbreviation',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'ancestors',
            'username',
            'full_name',
            'enabled',
            'password_modifiable',
            'require_new_password',
            'standard_auth',
            'memberships',
            'acg',
            'acl',
            'trust_id',
            'initials',
            'default_timezone',
            'ldapdn',
        ): users.get,
        'comments': objects_processors.get_info,
        'addresses': users.get_addresses,
        'security_roles': users.get_security_roles,
        'privileges': users.get_privileges,
        ('default_timezone', 'language'): users.get_settings,
    }
    _API_PATCH: dict = {
        (
            'name',
            'abbreviation',
            'username',
            'full_name',
            'enabled',
            'password',
            'description',
            'password_modifiable',
            'require_new_password',
            'password_expiration_date',
            'standard_auth',
            'ldapdn',
            'trust_id',
            'initials',
            'privileges',
            'memberships',
            'addresses',
            'security_roles',
            'default_timezone',
        ): (users.update, 'patch'),
        ('owner', 'comments'): (objects_processors.update, 'partial_put'),
        ('default_timezone', 'language'): (users.update_user_settings, 'partial_put'),
    }

    @staticmethod
    def _parse_owner(
        source: dict, connection: 'Connection', to_snake_case: bool = True
    ):
        """Parses owner from the API response."""
        from mstrio.users_and_groups import User

        return User.from_dict(source, connection, to_snake_case)

    @staticmethod
    def _parse_language(
        source: dict, connection: 'Connection', to_snake_case: bool = True
    ):
        """Parses language from the API response."""
        if not source['value']:
            return None
        source['id'] = source.pop('value')

        return Language.from_dict(source, connection, to_snake_case)

    @staticmethod
    def _parse_addresses(
        source: list[dict], connection: 'Connection', to_snake_case: bool = True
    ):
        """Parse Addresses from the API response."""
        from mstrio.users_and_groups.contact import ContactAddress

        return [
            ContactAddress.from_dict(contact, connection, to_snake_case)
            for contact in source
        ]

    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'password_expiration_date': time_helper.DatetimeFormats.FULLDATETIME,
        'language': _parse_language,
        'owner': _parse_owner,
        'addresses': _parse_addresses,
    }

    def __init__(
        self,
        connection: Connection,
        username: str | None = None,
        name: str | None = None,
        id: str | None = None,
    ) -> None:
        """Initialize User object by passing username, name, or id.
        When `id` is provided (not `None`), `username` and `name` are omitted.
        When `id` is not provided (`None`) and `username` is provided (not
        `None`), `name` is omitted.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`
            id: ID of User
            username: username of User
            name: name of User
        """
        if id is None and name is None and username is None:
            helper.exception_handler(
                "Please specify either 'id', 'username' or 'name' parameter in the "
                "constructor."
            )

        if id is None:
            users = (
                User._get_user_ids(
                    connection=connection,
                    abbreviation_begins=username,
                    abbreviation=username,
                )
                if username is not None
                else User._get_user_ids(
                    connection=connection,
                    name_begins=name,
                    name=name,
                )
            )

            if users:
                [id] = users
            else:
                temp_name = name if name else username
                helper.exception_handler(
                    f"There is no user: '{temp_name}'", exception_type=ValueError
                )

        super().__init__(
            connection=connection, object_id=id, username=username, name=name
        )

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.username = kwargs.get("username")
        self.full_name = kwargs.get("full_name")
        self.enabled = kwargs.get("enabled")
        self.password_modifiable = kwargs.get("password_modifiable")
        self.standard_auth = kwargs.get("standard_auth")
        self.require_new_password = kwargs.get("require_new_password")
        self.ldapdn = kwargs.get("ldapdn")
        self.trust_id = kwargs.get("trust_id")
        self._initials = kwargs.get("initials")
        self._addresses = kwargs.get("addresses")
        self._memberships = kwargs.get("memberships")
        self._security_roles = kwargs.get("security_roles")
        self._privileges = kwargs.get("privileges")
        self._security_filters = kwargs.get("security_filters")
        self._default_timezone = kwargs.get("default_timezone")
        self._language = kwargs.get("language")

    @classmethod
    def create(
        cls,
        connection: Connection,
        username: str,
        full_name: str,
        password: str | None = None,
        description: str | None = None,
        enabled: bool = True,
        password_modifiable: bool = True,
        password_expiration_date: str | datetime | None = None,
        require_new_password: bool = True,
        standard_auth: bool = True,
        ldapdn: str | None = None,
        trust_id: str | None = None,
        database_auth_login: str | None = None,
        memberships: list | None = None,
        language: 'str | Language | None' = None,
        default_timezone: 'str| dict | None' = None,
        owner: 'User | str | None' = None,
    ) -> "User":
        """Create a new user on the I-Server. Returns User object.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`
            username: username of user
            full_name: user full name
            password: user password
            description: user description
            enabled: specifies if user is allowed to log in
            password_modifiable: Specifies if user password can be modified
            password_expiration_date: Expiration date of user password either
                as a datetime or string: "yyyy-MM-dd HH:mm:ss" in UTC
            require_new_password: Specifies if user is required to provide a new
                password.
            standard_auth: Specifies whether standard authentication is allowed
                for user.
            ldapdn: User's LDAP distinguished name
            trust_id: Unique user ID provided by trusted authentication provider
            database_auth_login: Database Authentication Login
            memberships: specify User Group IDs which User will be member off.
            owner: owner of user
            default_timezone: default timezone for user
            language: language for user
        """
        password_expiration_date = time_helper.map_datetime_to_str(
            name='password_expiration_date',
            date=password_expiration_date,
            string_to_date_map=cls._FROM_DICT_MAP,
        )
        body = {
            "username": username,
            "fullName": full_name,
            "description": description,
            "password": password,
            "enabled": enabled,
            "passwordModifiable": password_modifiable,
            "passwordExpirationDate": password_expiration_date,
            "requireNewPassword": require_new_password,
            "standardAuth": standard_auth,
            "ldapdn": ldapdn,
            "trustId": trust_id,
            "databaseAuthLogin": database_auth_login,
            "memberships": memberships,
        }
        body = helper.delete_none_values(body, recursion=True)
        response = users.create(connection, body, username)
        if config.verbose:
            logger.info(
                f"Successfully created user named: '{response.get('username')}' "
                f"with ID: '{response.get('id')}'"
            )
        # User objects are created with default language and timezone
        # in order to change them, you have to go through separate API calls
        user_created = cls.from_dict(source=response, connection=connection)
        user_created.alter(
            language=language, default_timezone=default_timezone, owner=owner
        )
        return user_created

    @classmethod
    def _create_users_from_csv(
        cls, connection: Connection, csv_file: str
    ) -> list["User"]:
        func = cls.create
        args = helper.get_args_from_func(func)
        df = read_csv(csv_file, na_filter=False, usecols=lambda x: x in args)
        user_list = []
        all_param_value_dict = df.to_dict('records')

        for param_value_dict in all_param_value_dict:
            param_value_dict['connection'] = connection
            try:
                temp_user = func(**param_value_dict)
            except HTTPError:
                pass
            else:
                user_list.append(temp_user)

        return user_list

    @classmethod
    def _get_users(
        cls,
        connection: Connection,
        name_begins: str | None = None,
        abbreviation_begins: str | None = None,
        to_dictionary: bool = False,
        limit: int | None = None,
        **filters,
    ) -> list["User"] | list[dict]:
        msg = "Error getting information for a set of users."
        objects = users.get_all(
            connection=connection,
            limit=limit,
            msg=msg,
            name_begins=name_begins,
            abbreviation_begins=abbreviation_begins,
            filters=filters,
        )
        if to_dictionary:
            return objects
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]

    @classmethod
    def _get_user_ids(
        cls,
        connection: Connection,
        name_begins: str | None = None,
        abbreviation_begins: str | None = None,
        limit: int | None = None,
        **filters,
    ) -> list[str]:
        user_dicts = User._get_users(
            connection=connection,
            name_begins=name_begins,
            abbreviation_begins=abbreviation_begins,
            to_dictionary=True,
            limit=limit,
            **dict(filters),
        )
        return [user['id'] for user in user_dicts]

    def alter(
        self,
        username: str | None = None,
        full_name: str | None = None,
        description: str | None = None,
        password: str | None = None,
        enabled: bool | None = None,
        password_modifiable: bool | None = None,
        password_expiration_date: str | None = None,
        standard_auth: bool | None = None,
        require_new_password: bool | None = None,
        ldapdn: str | None = None,
        trust_id: str | None = None,
        database_auth_login: str | None = None,
        owner: 'User | str | None' = None,
        default_timezone: 'str| dict | None' = None,
        language: 'str | Language | None' = None,
        comments: str | None = None,
    ) -> None:
        """Alter user properties.

        Args:
            username: username of user
            full_name: user full name
            description: user description
            password: user password
            enabled: specifies if user is allowed to log in
            password_modifiable: Specifies if user password can be modified
            password_expiration_date: Expiration date of user password,
                "yyyy-MM-dd HH:mm:ss" in UTC
            standard_auth: Specifies whether standard authentication is allowed
                for user.
            require_new_password: Specifies if user is required to provide a new
                password.
            ldapdn: User's LDAP distinguished name
            trust_id: Unique user ID provided by trusted authentication provider
            database_auth_login: Database Authentication Login
            owner: owner of user
            default_timezone: default timezone for user
            language: language for user
            comments: long description of the object
        """
        owner = owner.id if isinstance(owner, User) else owner
        if language:
            language = {
                'value': (language.id if isinstance(language, Language) else language)
            }
        if default_timezone:
            default_timezone = {
                'id': (
                    default_timezone['id']
                    if isinstance(default_timezone, dict)
                    else default_timezone
                )
            }
        properties = filter_params_for_func(self.alter, locals(), exclude=['self'])
        self._alter_properties(**properties)

    def _create_address_v1(self, name, address, default):
        body = {
            "name": name,
            "deliveryMode": "EMAIL",
            "device": "GENERIC_EMAIL",
            "value": address,
            "default": default,
        }
        users.create_address(self.connection, self.id, body)
        response = users.get_addresses(self.connection, self.id)
        return response

    def _create_address_v2(self, name, address, default, delivery_type, device_id):
        if delivery_type and not device_id and config.verbose:
            msg = (
                f"Could not add address for user '{self.name}'. When you "
                "specify 'delivery_type', you need to add 'device_id' as "
                "well."
            )
            helper.exception_handler(msg, TypeError)
        if device_id and not delivery_type and config.verbose:
            msg = (
                f"Could not add address for user '{self.name}'. When you "
                "specify 'device_id', you need to add 'delivery_type' as "
                "well."
            )
            helper.exception_handler(msg, TypeError)

        body = {
            "name": name,
            "physicalAddress": address,
            "deliveryType": delivery_type,
            "deviceId": device_id,
            "idDefault": default,
        }
        response = users.create_address_v2(self.connection, self.id, body)
        return response

    def add_address(
        self,
        name: str | None = None,
        address: str | None = None,
        default: bool = True,
        delivery_type: str | None = None,
        device_id: str | None = None,
    ) -> None:
        """Add new address to the user object.

        Args:
            name (str): User-specified name for the address
            address (str): The actual value of the address i.e. email address
                associated with this address name/id
            default (bool, optional): Specifies whether this address is the
                default address (change isDefault parameter). Default value is
                set to True.
            delivery_type (str, optional): Delivery type
            device_id (str, optional): Device ID
        """
        if delivery_type or device_id:
            response = self._create_address_v2(
                name=name,
                address=address,
                default=default,
                delivery_type=delivery_type,
                device_id=device_id,
            )
        else:
            response = self._create_address_v1(
                name=name, address=address, default=default
            )
        if response:
            if config.verbose:
                logger.info(f"Added address '{address}' for user '{self.name}'")
            self._addresses = self._parse_addresses(
                response.get('addresses'), self.connection
            )

    def update_address(
        self,
        id: str,
        name: str | None = None,
        address: str | None = None,
        default: bool | None = None,
        delivery_type: str | None = None,
        device_id: str | None = None,
    ) -> None:
        """Update existing address. The address ID has to be specified
        as the name is not unique.

        Args:
            id (str): ID of the address
            name (str, optional): New user-specified name for the address
            address (str, optional): New address value
            default (bool, optional): Whether the address should be (un)marked
                as default
            delivery_type (str, optional): Delivery type
            device_id (str, optional): Device ID
        """
        if not id:
            helper.exception_handler("Please specify 'id' parameter in the method.")

        address_for_update = helper.filter_list_of_dicts(
            [address.to_dict() for address in self.addresses], id=id
        )[0]

        body = {
            'name': name if name else address_for_update['name'],
            'physicalAddress': (
                address if address else address_for_update['physicalAddress']
            ),
            'deliveryType': (
                delivery_type if delivery_type else address_for_update['deliveryType']
            ),
            'deviceId': device_id if device_id else address_for_update['deviceId'],
            'isDefault': default if default else address_for_update['isDefault'],
        }
        if users.update_address(self.connection, self.id, id, body):
            if config.verbose:
                logger.info(f"Updated address with ID '{id}' for user '{self.name}'")
            self.fetch("addresses")

    def remove_address(
        self,
        name: str | None = None,
        address: str | None = None,
        id: str | None = None,
    ) -> None:
        """Remove existing address from the user object. Specify either address
        ID or name. Warning, address names are not unique and can potentially
        remove multiple addresses.

        Args:
            name: User-specified name for the address
            address: The actual value of the address i.e. email address
                associated with this address name/id
            id: ID of the address.
        """
        initial_addresses = [address.to_dict() for address in self.addresses]
        if name is None and id is None and address is None or (name and id and address):
            helper.exception_handler(
                "Please specify either 'name' or 'id' parameter in the method."
            )
        if id is not None:
            addresses = helper.filter_list_of_dicts(initial_addresses, id=id)
            new_addresses = helper.filter_list_of_dicts(initial_addresses, id=f'!={id}')
        elif address is not None:
            addresses = helper.filter_list_of_dicts(initial_addresses, value=address)
            new_addresses = helper.filter_list_of_dicts(
                initial_addresses, value=f'!={address}'
            )
        elif name is not None:
            addresses = helper.filter_list_of_dicts(initial_addresses, name=name)
            new_addresses = helper.filter_list_of_dicts(
                initial_addresses, name=f'!={name}'
            )

        for addr in addresses:
            response = users.delete_address(
                self.connection, id=self.id, address_id=addr['id']
            )
            if response:
                if config.verbose:
                    logger.info(
                        f"Removed address '{addr['name']}' with id {addr['id']} "
                        f"from user '{self.name}'"
                    )
                self._addresses = self._parse_addresses(new_addresses, self.connection)

    def add_to_user_groups(
        self, user_groups: "str | UserGroup | list[str | UserGroup]"
    ) -> None:
        """Adds this User to user groups specified in user_groups.

        Args:
            user_groups: list of `UserGroup` objects or IDs
        """
        succeeded, failed = self._update_nested_properties(
            user_groups, "memberships", "add"
        )
        if succeeded and config.verbose:
            logger.info(f"Added user '{self.name}' to group(s): {succeeded}")
        elif failed and config.verbose:
            logger.info(f"User '{self.name}' is already a member of {failed}")

    def remove_from_user_groups(
        self, user_groups: "str | UserGroup | list[str | UserGroup]"
    ) -> None:
        """Removes this User from user groups specified in user_groups.

        Args:
            user_groups: list of `UserGroup` objects or IDs
        """
        succeeded, failed = self._update_nested_properties(
            user_groups, "memberships", "remove"
        )
        if succeeded and config.verbose:
            logger.info(f"Removed user '{self.name}' from group(s): {succeeded}")
        elif failed and config.verbose:
            logger.info(f"User '{self.name}' is not in {failed} group(s)")

    def remove_from_all_user_groups(self) -> None:
        """Removes this User from all user groups."""
        memberships = self.memberships
        existing_ids = [obj.get('id') for obj in memberships]
        self.remove_from_user_groups(user_groups=existing_ids)

    def assign_security_role(
        self, security_role: SecurityRole | str, project: 'Project | str'
    ) -> None:  # NOSONAR
        """Assigns a Security Role to the user for given project.

        Args:
            security_role: Security Role ID or object
            project: Project name or object
        """

        security_role = (
            security_role
            if isinstance(security_role, SecurityRole)
            else SecurityRole(self.connection, id=str(security_role))
        )

        security_role.grant_to([self.id], project)
        if config.verbose:
            logger.info(
                f"Assigned Security Role '{security_role.name}' to user: '{self.name}'"
            )
        self.fetch('security_roles')

    def revoke_security_role(
        self, security_role: SecurityRole | str, project: 'Project | str'
    ) -> None:  # NOSONAR
        """Removes a Security Role from the user for given project.

        Args:
            security_role: Security Role ID or object
            project: Project name or object
        """

        security_role = (
            security_role
            if isinstance(security_role, SecurityRole)
            else SecurityRole(self.connection, id=str(security_role))
        )

        security_role.revoke_from([self.id], project)
        if config.verbose:
            logger.info(
                f"Revoked Security Role '{security_role.name}' from user: '{self.name}'"
            )
        self.fetch('security_roles')

    @method_version_handler('11.3.0200')
    def list_security_filters(
        self, projects: str | list[str] | None = None, to_dictionary: bool = False
    ) -> dict:
        """Get the list of security filters for user. They can be filtered by
        the projects' ids.

        Args:
            projects (str or list of str, optional): collection of projects' ids
                which is used for filtering data
            to_dictionary: If True returns security filters as dicts, by default
                (False) returns them as objects.

        Returns:
            Dictionary with project names as keys and list with security
            filters as values. In case of no security filter for the given user
            in the particular project, then this project is not placed in
            the dictionary.
        """
        from mstrio.modeling.security_filter import SecurityFilter

        objects_ = users.get_security_filters(self.connection, self.id, projects)
        projects_ = objects_.get("projects")

        objects_ = {
            project_.get("name"): project_.get("securityFilters")
            for project_ in projects_
            if project_.get("securityFilters")
        }

        security_filters = {
            name: [
                SecurityFilter.from_dict(sec_filter, self.connection)
                for sec_filter in sec_filters
            ]
            for (name, sec_filters) in objects_.items()
        }
        if to_dictionary:
            return objects_
        return security_filters

    def apply_security_filter(self, security_filter: "SecurityFilter | str") -> bool:
        """Apply a security filter to the user.

        Args:
            security_filter (string or object): identifier of security filter or
                `SecurityFilter` object which will be applied to the user.
        Returns:
            True when applying was successful. False otherwise.
        """
        if isinstance(security_filter, str):
            from mstrio.modeling.security_filter import SecurityFilter

            security_filter = SecurityFilter.from_dict(
                {"id": security_filter}, self.connection
            )
        return security_filter.apply(self.id)

    def revoke_security_filter(self, security_filter: "SecurityFilter | str") -> bool:
        """Revoke a security filter from the user.

        Args:
            security_filter (string or object): identifier of security filter or
                `SecurityFilter` object which will be revoked from the user.

        Returns:
            True when revoking was successful. False otherwise.
        """
        if isinstance(security_filter, str):
            from mstrio.modeling.security_filter import SecurityFilter

            security_filter = SecurityFilter.from_dict(
                {"id": security_filter}, self.connection
            )
        return security_filter.revoke(self.id)

    def grant_privilege(
        self, privilege: "str | list[str] | Privilege | list[Privilege]"
    ) -> None:
        """Grant privileges directly to the user.

        Args:
            privilege: list of privilege objects, ids or names
        """
        from mstrio.access_and_security.privilege import Privilege

        privileges = [
            priv['id']
            for priv in Privilege._validate_privileges(self.connection, privilege)
        ]
        existing_ids = [
            privilege['privilege']['id']
            for privilege in self.list_privileges(mode='GRANTED')
        ]
        succeeded, failed = self._update_nested_properties(
            privileges, "privileges", "add", existing_ids
        )

        if succeeded:
            self.fetch(
                'privileges'
            )  # fetch the object properties and set object attributes
            if config.verbose:
                logger.info(f"Granted privilege(s) {succeeded} to '{self.name}'")
        if failed and config.verbose:
            logger.info(f"User '{self.name}' already has privilege(s) {failed}")

    def revoke_privilege(
        self, privilege: "str | list[str] | Privilege | list[Privilege]"
    ) -> None:
        """Revoke directly granted user privileges.

        Args:
            privilege: list of privilege objects, ids or names
        """
        from mstrio.access_and_security.privilege import Privilege

        privileges = {
            priv['id']
            for priv in Privilege._validate_privileges(self.connection, privilege)
        }
        existing_ids = [
            privilege['privilege']['id']
            for privilege in self.list_privileges(mode='ALL')
        ]
        directly_granted = {
            privilege['privilege']['id']
            for privilege in self.list_privileges(mode='GRANTED')
        }
        to_revoke = list(privileges.intersection(directly_granted))
        not_directly_granted = list(
            (set(existing_ids) - directly_granted).intersection(privileges)
        )

        if not_directly_granted:
            msg = (
                f"Privileges {sorted(not_directly_granted)} are inherited and will be "
                "omitted. Only directly granted privileges can be revoked by this "
                "method."
            )
            helper.exception_handler(msg, exception_type=Warning)

        succeeded, failed = self._update_nested_properties(
            to_revoke, "privileges", "remove", existing_ids
        )
        if succeeded:
            self.fetch(
                'privileges'
            )  # fetch the object properties and set object attributes
            if config.verbose:
                logger.info(f"Revoked privilege(s) {succeeded} from '{self.name}'")
        if failed and config.verbose:
            logger.warning(f"User '{self.name}' does not have privilege(s) {failed}")

    def revoke_all_privileges(self, force: bool = False) -> None:
        """Revoke directly granted user privileges.

        Args:
            force: If True, no additional prompt will be shown before revoking
                all privileges from User.
        """
        user_input = 'N'
        if not force:
            user_input = input(
                f"Are you sure you want to revoke all privileges from user "
                f"'{self.name}'? [Y/N]: "
            )
        if force or user_input == 'Y':
            to_revoke = [
                privilege['privilege']['id']
                for privilege in self.list_privileges(mode='GRANTED')
            ]
            if to_revoke:
                self.revoke_privilege(privilege=to_revoke)
            else:
                logger.info(
                    f"User '{self.name}' does not have any directly granted privileges"
                )

    def list_privileges(
        self, mode: PrivilegeMode | str = PrivilegeMode.ALL, to_dataframe: bool = False
    ) -> list:
        """List privileges for user.

        Args:
            mode: Filter by source of privileges. One of: `ALL`, `INHERITED`,
                or `GRANTED`. See: `privilege.PrivilegeMode` enum.
            to_dataframe: If True, return a `DataFrame` object containing
                privileges.
        """
        self.fetch('privileges')

        def to_df(priv_list):
            priv_dict = {}
            for priv in priv_list:
                priv_dict[priv['privilege']['id']] = priv['privilege']['name']
            df = DataFrame.from_dict(priv_dict, orient='index', columns=['Name'])
            df.index.name = 'ID'
            return df

        if not isinstance(mode, PrivilegeMode):
            try:
                mode = PrivilegeMode(mode)
            except ValueError:
                msg = (
                    "Wrong privilege mode has been passed, allowed modes are ['ALL'/"
                    "'INHERITED'/'GRANTED']. See: `privilege.PrivilegeMode` enum."
                )
                helper.exception_handler(msg, ValueError)

        privileges = []
        if mode == PrivilegeMode.ALL:
            privileges = self.privileges
        elif mode == PrivilegeMode.INHERITED:
            for privilege in self.privileges:
                is_inherited = False
                for source in privilege['sources']:
                    is_inherited = not source['direct'] or is_inherited
                if is_inherited:
                    privileges.append(privilege)
        else:  # GRANTED
            for privilege in self.privileges:
                is_granted = False
                for source in privilege['sources']:
                    is_granted = source['direct'] or is_granted
                if is_granted:
                    privileges.append(privilege)

        return to_df(privileges) if to_dataframe else privileges

    def disconnect(self, nodes: str | list[str] | None = None) -> None:
        """Disconnect all active user connection sessions for the specified
        node.

        Args:
            nodes: list of node names
        """
        temp_connections = UserConnections(self.connection)
        temp_connections.disconnect_users(users=self, nodes=nodes)

    def delete(self, force: bool = False) -> bool:
        """Deletes the user.
        Deleting the user will not remove the user's shared files.

                Args:
            force: If True, no additional prompt will be shown before deleting
                User.
        Returns:
            True for success. False otherwise.
        """
        return super().delete(force=force)

    def _to_dataframe_as_columns(
        self, properties: list[str] | None = None
    ) -> pd.DataFrame:
        """Exports user object to dataframe, with properties as columns

        Args:
            properties (list, optional): list of properties to be exported

        Returns: dataframe
        """
        excluded_keys = ['connection']
        selected_keys = properties or (self.list_properties().keys() - excluded_keys)

        def convert(obj, inside=False):
            if isinstance(obj, IntFlag):
                return obj.value
            if isinstance(obj, (str, int)):
                return obj
            if isinstance(obj, Enum):
                return obj.value
            if isinstance(obj, datetime):
                return str(obj)

            result = None

            if isinstance(obj, list):
                result = [convert(el, inside=True) for el in obj]
            if isinstance(obj, Dictable):
                result = dict(sorted(obj.to_dict().items()))
            if isinstance(obj, dict):
                result = {
                    key: convert(value, inside=True) for key, value in obj.items()
                }

            return result if inside else json.dumps(result)

        data = {
            key: convert(value)
            for key, value in self.list_properties().items()
            if key in selected_keys
        }
        return pd.DataFrame(data, index=[0])

    @classmethod
    def to_datafame_from_list(
        cls, objects: list['User'], properties: list[str] | None = None
    ) -> pd.DataFrame:
        """Exports list of user objects to dataframe.
        The properties that are lists, dictionaries, or objects of
        custom classes, are first converted to dictionary using `to_dict()`
        method, then serialized string as json.

        Args:
            objects (list): list of user objects to be exported
            properties (list, optional): list of properties of user object
                to be exported

        Returns: dataframe

        Raises:
            TypeError if `objects` parameter contains objects other than
             of type User
        """
        dataframes = []

        for obj in objects:
            if not isinstance(obj, User):
                raise TypeError('All objects in the list must be of type User.')

            dataframes.append(obj._to_dataframe_as_columns(properties))

        return pd.concat(dataframes, ignore_index=True)

    @classmethod
    def to_csv_from_list(
        cls,
        objects: list['User'],
        path: str | None = None,
        properties: list[str] | None = None,
    ) -> str | None:
        """Exports list of user objects to csv (if path is provided)
        or to string (if path is not provided).
        The properties that are lists, dictionaries, or objects of
        custom classes, are first converted to dictionary using `to_dict()`
        method, then serialized string as json.

        Args:
            objects (list): list of user objects to be exported
            path (str, optional): a path specifying where to save the csv file
            properties (list, optional): list of properties of user object
                to be exported

        Returns: str or None

        Raises:
            TypeError if `objects` parameter contains objects other than
             of type User
        """
        dataframe = cls.to_datafame_from_list(objects, properties)
        return dataframe.to_csv(index=False, path_or_buf=path)

    @property
    def memberships(self):
        return self._memberships

    @property
    def initials(self):
        return self._initials

    @property
    def addresses(self):
        return self._addresses

    @property
    def security_roles(self):
        return self._security_roles

    @property
    def privileges(self):
        self.fetch('privileges')
        return self._privileges

    @property
    @method_version_handler('11.3.0500')
    def default_timezone(self):
        return self._default_timezone

    @property
    @method_version_handler('11.3.0700')
    def language(self):
        return self._language

    @property
    def security_filters(self):
        if not self._security_filters:
            self._security_filters = self.list_security_filters()
        return self._security_filters



import logging

from packaging import version

from mstrio import config
from mstrio.api import events
from mstrio.connection import Connection
from mstrio.utils import helper
from mstrio.utils.entity import DeleteMixin, Entity, ObjectTypes
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.translation_mixin import TranslationMixin
from mstrio.utils.version_helper import class_version_handler, method_version_handler

logger = logging.getLogger(__name__)


@method_version_handler('11.3.0100')
def list_events(
    connection: Connection, to_dictionary: bool = False, limit: int = None, **filters
) -> list["Event"] | list[dict]:
    """List event objects or event dictionaries. Optionally filter list.

    Args:
        connection(object): MicroStrategy connection object returned
            by 'connection.Connection()'
        to_dictionary(bool, optional): if True, return event as
            list of dicts
        limit(int, optional): maximum number of Events returned.
        **filters: Available filter parameters:
            ['name', 'id', 'description', 'acg']
    """
    _objects = helper.fetch_objects(
        connection=connection,
        api=events.list_events,
        limit=limit,
        filters=filters,
        dict_unpack_value='events',
    )

    if to_dictionary:
        return _objects
    else:
        return [Event.from_dict(source=obj, connection=connection) for obj in _objects]


@class_version_handler('11.3.0100')
class Event(Entity, DeleteMixin, TranslationMixin):
    """Class representation of MicroStrategy Event object.

    Attributes:
        connection: A MicroStrategy connection object
        name: Event name
        id: Event ID
        description: Event descriptions
    """

    _PATCH_PATH_TYPES = {'name': str, 'description': str}
    _OBJECT_TYPE = ObjectTypes.SCHEDULE_EVENT
    _API_GETTERS = {
        (
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'comments',
        ): objects_processors.get_info,
    }
    _API_DELETE = staticmethod(events.delete_event)
    _API_PATCH = {
        ('name', 'description'): (events.update_event, 'put'),
        'comments': (objects_processors.update, 'partial_put'),
    }

    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        name: str | None = None,
    ) -> None:
        """Initialize the Event object, populates it with I-Server data.
        Specify either `id` or `name`. When `id` is provided (not `None`),
        `name` is omitted.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`.
            id: Event ID
            name: Event name
        """
        self._API_GETTERS[('id', 'name', 'description')] = (
            events.get_event
            if version.parse(connection.web_version) >= version.parse('11.3.0200')
            else objects_processors.get_info
        )

        if id is None and name is None:
            raise AttributeError(
                "Please specify either 'name' or 'id' parameter in the constructor."
            )
        if id is None:
            objects_info = list_events(connection, name=name, to_dictionary=True)
            if objects_info:
                object_info, object_info["connection"] = objects_info[0], connection
                self._init_variables(**object_info)
            else:
                raise ValueError(f"There is no event with the given name: '{name}'")
        else:
            super().__init__(connection=connection, object_id=id, name=name)

    def trigger(self):
        """Trigger the Event"""
        response = events.trigger_event(self.connection, self.id)
        if response.ok and config.verbose:
            logger.info(
                f"Event '{self.name}' with ID : '{self.id}' has been triggered."
            )
        return response.ok

    @classmethod
    def create(
        cls, connection: Connection, name: str, description: str | None = None
    ) -> "Event":
        """Create an Event

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`.
            name: Name of the new Event
            description: Description of the new Event
        """
        body = helper.delete_none_values(
            {
                "name": name,
                "description": description,
            },
            recursion=True,
        )
        response = events.create_event(connection, body)
        return cls.from_dict(response.json(), connection)

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        comments: str | None = None,
    ):
        """Alter the Event's properties

        Args:
            name: New name for the Event
            description: New description for the Event
            comments: long description of the Event
        """
        args = helper.delete_none_values(
            {
                "name": name,
                "description": description,
            },
            recursion=True,
        )
        self._alter_properties(**args)
        if config.verbose:
            logger.info(f"Updated subscription '{self.name}' with ID: {self.id}.")



import json
from typing import TYPE_CHECKING
from unittest.mock import Mock
from urllib.parse import quote, urlencode

from packaging import version

from mstrio.helpers import MstrException, PartialSuccess, Success
from mstrio.utils.error_handlers import ErrorHandler, bulk_operation_response_handler
from mstrio.utils.helper import (
    delete_none_values,
    filter_list_of_dicts,
    response_handler,
)
from mstrio.utils.sessions import FuturesSessionWithRenewal

if TYPE_CHECKING:
    from requests import Response

    from mstrio.connection import Connection

ISERVER_VERSION_11_3_2 = '11.3.0200'


@ErrorHandler(err_msg='Error getting list of all projects from metadata.')
def get_projects(
    connection: 'Connection',
    offset: int = 0,
    limit: int = -1,
    error_msg: str | None = None,
):
    """Get list of all projects from metadata.
    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        offset(int): Starting point within the collection of returned search
            results. Used to control paging behavior.
        limit(int): Maximum number of items returned for a single search
            request. Used to control paging behavior. Use -1 (default ) for no
            limit (subject to governing settings).
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """

    return connection.get(
        endpoint='/api/monitors/projects',
        headers={'X-MSTR-ProjectID': None},
        params={'offset': offset, 'limit': limit},
    )


def get_projects_async(
    future_session: 'FuturesSessionWithRenewal', offset: int = 0, limit: int = -1
):
    """Get list of all projects from metadata asynchronously.
    Args:
        future_session(object): `FuturesSessionWithRenewal` object to call
            MicroStrategy REST Server asynchronously
        offset(int): Starting point within the collection of returned search
            results. Used to control paging behavior.
        limit(int): Maximum number of items returned for a single search
            request. Used to control paging behavior. Use -1 (default ) for no
            limit (subject to governing settings).

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    endpoint = '/api/monitors/projects'
    headers = {'X-MSTR-ProjectID': None}
    params = {'offset': offset, 'limit': limit}
    future = future_session.get(endpoint=endpoint, headers=headers, params=params)
    return future


@ErrorHandler(
    err_msg='Error getting information about nodes in the connected Intelligence '
    'Server cluster.'
)
def get_node_info(
    connection: 'Connection',
    id: str | None = None,
    node_name: str | None = None,
    error_msg: str | None = None,
):
    """Get information about nodes in the connected Intelligence Server
    cluster.

    This includes basic information, runtime state and information of projects
    on each node. This operation requires the "Monitor cluster" privilege.

    Args:
        connection(object): MicroStrategy connection object returned by

        id (str, optional): Project ID
        node_name (str, optional): Node Name
        error_msg (string, optional): Custom Error Message for Error Handling
    """
    return connection.get(
        endpoint='/api/monitors/iServer/nodes',
        headers={'X-MSTR-ProjectID': None},
        params={'projects.id': id, 'name': node_name},
    )


@ErrorHandler(
    err_msg='Error updating properties for a project {project_id} for cluster node {'
    'node_name}.'
)
def update_node_properties(
    connection: 'Connection',
    node_name: str,
    project_id: str,
    body: dict,
    error_msg: str | None = None,
    whitelist: list[tuple] | None = None,
):
    """Update properties such as project status for a specific project for
    respective cluster node. You obtain cluster node name and project id from
    GET /monitors/iServer/nodes.

    {
        'operationList': [
            {
                'op': 'replace',
                'path': '/status',
                'value': 'loaded'
            }
        ]
    }

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        node_name (string): Node Name.
        project_id (string): Project ID.
        body (JSON): Body 'op' can have "value" set to "add", "replace",
            "remove"; 'path' can have pattern: /([/A-Za-z0-9~])*-* example:
            /status; 'values' for '/status'  we can choose [loaded, unloaded,
            request_idle, exec_idle, wh_exec_idle, partial_idle, full_idle]
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """

    return connection.patch(
        endpoint=f'/api/monitors/iServer/nodes/{node_name}/projects/{project_id}',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(
    err_msg='Error adding node {node_name} to connected Intelligence Server cluster.'
)
def add_node(
    connection: 'Connection',
    node_name: str,
    error_msg: str | None = None,
    whitelist: list[tuple] | None = None,
):
    """Add a node to the connected Intelligence Server cluster. The node must
    meet I-Server clustering requirements. If the node is part of a multi-node
    cluster, all the nodes in that cluster will be added. If the node is
    already in the cluster, the operation succeeds without making any change.
    This operation requires the "Monitor cluster" and "Administer cluster"
    privilege.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        node_name (string): Node Name.
        error_msg (string, optional): Custom Error Message for Error Handling
        whitelist(list): list of tuples of I-Server Error and HTTP errors codes
            respectively, which will not be handled
            i.e. whitelist = [('ERR001', 500),('ERR004', 404)]

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """

    return connection.put(
        endpoint=f'/api/monitors/iServer/nodes/{node_name}',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(
    err_msg='Error removing node {node_name} from the connected Intelligence Server '
    'cluster.'
)
def remove_node(
    connection: 'Connection',
    node_name: str,
    error_msg: str | None = None,
    whitelist: list[tuple] | None = None,
):
    """Remove a node from the connected Intelligence Server cluster. After a
    successful removal, some existing authorization tokens may become
    invalidated and in this case re-login is needed. You cannot remove a node
    if it's the configured default node of Library Server, or there is only one
    node in the cluster. This operation requires the "Monitor cluster" and
    "Administer cluster" privilege.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        node_name (string): Node Name.
        error_msg (string, optional): Custom Error Message for Error Handling
        whitelist(list): list of tuples of I-Server Error and HTTP errors codes
            respectively, which will not be handled
            i.e. whitelist = [('ERR001', 500),('ERR004', 404)]

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """

    return connection.delete(
        endpoint=f'/api/monitors/iServer/nodes/{node_name}',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg='Error getting user connections for {node_name} cluster node.')
def get_user_connections(
    connection: 'Connection',
    node_name: str,
    offset: int = 0,
    limit: int = 100,
    error_msg: str | None = None,
):
    """Get user connections information on specific intelligence server node.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        offset(int): Starting point within the collection of returned search
            results. Used to control paging behavior.
        limit(int): Maximum number of items returned for a single search
            request. Used to control paging behavior. Use -1 (default ) for no
            limit (subject to governing settings).
        node_name (string): Node Name.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.get(
        endpoint='/api/monitors/userConnections',
        headers={'X-MSTR-ProjectID': None},
        params={'clusterNode': node_name, 'offset': offset, 'limit': limit},
    )


def get_user_connections_async(
    future_session: 'FuturesSessionWithRenewal',
    node_name: str,
    offset: int = 0,
    limit: int = 100,
):
    """Get user connections information on specific intelligence server node.

    Args:
        future_session(object): `FuturesSessionWithRenewal` object to call
            MicroStrategy REST Server asynchronously
        node_name (string): Node Name.
        offset(int): Starting point within the collection of returned search
            results. Used to control paging behavior.
        limit(int): Maximum number of items returned for a single search
            request. Used to control paging behavior. Use -1 (default ) for no
            limit (subject to governing settings).

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    params = {'clusterNode': node_name, 'offset': offset, 'limit': limit}
    endpoint = '/api/monitors/userConnections'
    headers = {'X-MSTR-ProjectID': None}
    future = future_session.get(endpoint=endpoint, headers=headers, params=params)
    return future


def delete_user_connection(
    connection: 'Connection', id: str, error_msg: str | None = None, bulk: bool = False
):
    """Disconnect a user connection on specific intelligence server node.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str, optional): Project ID
        error_msg (string, optional): Custom Error Message for Error Handling
        bulk(bool,optional): Whether to ignore errors when using request for
            bulk disconnect
    """
    response = connection.delete(
        endpoint=f'/api/monitors/userConnections/{id}',
        headers={'X-MSTR-ProjectID': None},
    )
    if not response.ok and bulk:
        if error_msg is None:
            error_msg = f"Error deleting user connections {id}."
        # whitelist error related to disconnecting yourself or other unallowed
        response_handler(response, error_msg, whitelist=[('ERR001', 500)])
    return response


def delete_user_connection_async(future_session: 'FuturesSessionWithRenewal', id: str):
    """Disconnect a user connection on specific intelligence server node.

    Args:
        future_session(object): `FuturesSessionWithRenewal` object to call
            MicroStrategy REST Server asynchronously
        id (str, optional): Project ID
    """

    endpoint = f'/api/monitors/userConnections/{id}'
    headers = {'X-MSTR-ProjectID': None}
    future = future_session.delete(endpoint=endpoint, headers=headers)
    return future


def delete_user_connections(connection: 'Connection', ids: list[str]):
    """Delete user connections on specific intelligence server node.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        ids (list of strings): list with ids of user connections to be deleted.

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    body = {'userConnectionIds': ids}
    response = connection.post(
        endpoint='/api/monitors/deleteUserConnections', json=body
    )
    return response


@ErrorHandler(err_msg='Error getting cube cache {id} info.')
def get_cube_cache_info(connection: 'Connection', id: str):
    """Get an single cube cache info.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`.
        id (string): cube cache id

    Returns:
        Complete HTTP response object.
    """
    return connection.get(endpoint=f'/api/monitors/caches/cubes/{id}')


@ErrorHandler(err_msg='Error deleting cube cache with ID {id}')
def delete_cube_cache(connection: 'Connection', id: str, throw_error: bool = True):
    """Delete an cube cache.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`.
        id (string): cube cache id
        throw_error (bool, optional): Flag indicates if the error should be
            thrown.

    Returns:
        Complete HTTP response object.
    """
    return connection.delete(endpoint=f'/api/monitors/caches/cubes/{id}')


@ErrorHandler(err_msg='Error altering cube cache {id} status.')
def alter_cube_cache_status(
    connection: 'Connection',
    id: str,
    active: bool | None = None,
    loaded: bool | None = None,
    throw_error: bool = True,
):
    """Alter an cube cache status. In one request it is possible to set either
    'active' or 'loaded', never both.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`.
        id (string): cube cache id,
        active (bool, optional): make cube cache active (True) or inactive
            (False)
        loaded (bool, optional): load (True) or unload (False) the cube cache
        throw_error (bool, optional): Flag indicates if the error should be
            thrown.

    Returns:
        Complete HTTP response object.
    """
    if loaded is not None:
        loaded = 'loaded' if loaded else 'unloaded'
    body = {'state': {'active': active, 'loadedState': loaded}}
    body = delete_none_values(body, recursion=True)

    return connection.patch(
        endpoint=f'/api/monitors/caches/cubes/{id}',
        headers={'Prefer': 'respond-async'},
        json=body,
    )


@ErrorHandler(err_msg='Error getting list of cube caches for node {node}.')
def get_cube_caches(
    connection: 'Connection',
    node: str,
    offset: int = 0,
    limit: int = 1000,
    project_ids: str | None = None,
    loaded: bool = False,
    sort_by: str | None = None,
    error_msg: str | None = None,
):
    """Get the list of cube caches on a specific cluster node.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`.
        node (string): Intelligence Server cluster node name
        offset (integer, optional): Starting point within the collection of
            returned results. Used to control paging behavior. Default value: 0.
        limit (integer, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default value: 1000.
        project_ids (string, optional): project id collection which is used for
            filtering data, for example:
            'B19DEDCC11D4E0EFC000EB9495D0F6E2,A232EDCC11D4E0EFC000EB9495D0F6E2'
        loaded (bool, optional): filter field which is used to filtering loaded
            cube cache. If True then the filter will be applied. Otherwise all
            cubes will be returned.
        sort_by (string, optional): Specify sorting criteria. For example
            '+name,-size' means sorting name ascending and size descending.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    loaded = 'loaded' if loaded else None
    return connection.get(
        endpoint='/api/monitors/caches/cubes',
        params={
            'clusterNode': node,
            'offset': offset,
            'limit': limit,
            'projectIds': project_ids,
            'state.loadedState': loaded,
            'sortBy': sort_by,
        },
    )


def get_cube_caches_async(
    future_session: 'FuturesSessionWithRenewal',
    node: str,
    offset: int = 0,
    limit: int = 1000,
    project_ids: str | None = None,
    loaded: bool = False,
    sort_by: str | None = None,
):
    """Get the list of cube caches on a specific cluster node asynchronously.

    Args:
        future_session(object): Future Session object to call MicroStrategy REST
            Server asynchronously
        node (string): Intelligence Server cluster node name
        offset (integer, optional): Starting point within the collection of
            returned results. Used to control paging behavior. Default value: 0.
        limit (integer, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default value: 1000.
        project_ids (string, optional): project id collection which is used for
            filtering data, for example:
            'B19DEDCC11D4E0EFC000EB9495D0F6E2,A232EDCC11D4E0EFC000EB9495D0F6E2'
        loaded (bool, optional): filter field which is used to filtering loaded
            cube cache. If True then the filter will be applied. Otherwise all
            cubes will be returned.
        sort_by (string, optional): Specify sorting criteria. For example
            '+name,-size' means sorting name ascending and size descending.
    Returns:
        Future with HTTP response returned by the MicroStrategy REST server as
        a result.
    """
    endpoint = '/api/monitors/caches/cubes'
    loaded = 'loaded' if loaded else None
    params = {
        'clusterNode': node,
        'offset': offset,
        'limit': limit,
        'projectIds': project_ids,
        'state.loadedState': loaded,
        'sortBy': sort_by,
    }
    future = future_session.get(endpoint=endpoint, params=params)
    return future


@ErrorHandler(err_msg='Error getting cube cache manipulation {manipulation_id} status.')
def get_cube_cache_manipulation_status(
    connection: 'Connection', manipulation_id: str, throw_error: bool = True
):
    """Get the manipulation status of cube cache.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`.
        manipulation_id (string): cube cache manipulation ID
        throw_error (bool, optional): In case of True (default) the error will
            be thrown when it occurs.

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/monitors/caches/cubes/manipulations/{manipulation_id}/status'
    )


@ErrorHandler(
    err_msg='Error getting database connections for {nodes_names} cluster node.'
)
def get_database_connections(
    connection: 'Connection', nodes_names: str, error_msg: str | None = None
):
    """Get database connections information on specific intelligence
        server node.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        nodes_names (string): Node names split by ",".
        error_msg (string, optional): Custom Error Message for Error Handling
    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    return connection.get(
        endpoint='/api/monitors/dbConnectionInstances',
        params={'clusterNodes': nodes_names},
    )


@ErrorHandler(err_msg='Error deleting database connections {connection_id}.')
def delete_database_connection(
    connection: 'Connection', connection_id: str, error_msg: str | None = None
):
    """Disconnect a database connection on specific intelligence server node.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        connection_id (str, optional): Database Connection Id
        error_msg (string, optional): Custom Error Message for Error Handling
    """
    return connection.delete(
        endpoint=f'/api/monitors/dbConnectionInstances/{connection_id}'
    )


def delete_database_connection_async(
    future_session: 'FuturesSessionWithRenewal', connection_id: str
):
    """Disconnect a database connection on specific intelligence server node.

    Args:
        future_session: Future Session object to call MicroStrategy REST
            Server asynchronously
        connection_id (str, optional): Database Connection Id
    """
    endpoint = f'/api/monitors/dbConnectionInstances/{connection_id}'
    return future_session.delete(endpoint=endpoint)


def get_job(
    connection: 'Connection',
    id: str,
    node_name: str | None = None,
    fields: list[str] | None = None,
    error_msg: str | None = None,
):
    """Get job information.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        node_name(str, optional): Node name, if not passed list jobs
            on all nodes
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (string, optional): Custom Error Message for Error Handling
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    response = Mock()  # create empty mock object to mimic REST API response

    if not node_name:
        # fetch jobs on all nodes
        nodes_response = get_node_info(connection).json()
        nodes = nodes_response['nodes']
        node_names = [node['name'] for node in nodes]

    if isinstance(node_name, str):
        node_names = [node_names]

    with FuturesSessionWithRenewal(connection=connection, max_workers=8) as session:
        futures = [
            get_jobs_async(future_session=session, node_name=node, fields=fields)
            for node in node_names
        ]
        jobs = []
        for f in futures:
            response = f.result()
            if not response.ok:
                response_handler(response, error_msg, throw_error=False)
            else:
                jobs.extend(response.json()['jobs'])

    job = filter_list_of_dicts(jobs, id=id)
    if not job:
        response.status_code = 400
        response.reason = f"Error getting job '{id}'"
        response.raise_for_status()
    elif len(job) > 1:
        response.status_code = 400
        response.reason = f"More than one job with id '{id}' was found."
        response.raise_for_status()
    else:
        job = job[0]
        job = json.dumps(job).encode('utf-8')
        response._content = job
        response.status_code = 200
        return response


@ErrorHandler(err_msg="Error getting job {id}.")
def get_job_v2(
    connection: 'Connection',
    id: str,
    fields: list[str] | None = None,
    error_msg: str | None = None,
):
    """Get job information.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (string, optional): Custom Error Message for Error Handling
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    return connection.get(
        endpoint=f'/api/v2/monitors/jobs/{id}',
        params={'fields': ','.join(fields) if fields else None},
    )


@ErrorHandler(err_msg="Error getting jobs list.")
def get_jobs(
    connection: 'Connection',
    node_name: str,
    project_id: str | None = None,
    status: str | None = None,
    job_type: str | None = None,
    user_full_name: str | None = None,
    object_id: str | None = None,
    sort_by: str | None = None,
    fields: list[str] | None = None,
    error_msg: str | None = None,
) -> 'Response':
    """Get list of a jobs.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        node_name(str): Node name,
        project_id(str, optional): Project id ,
        status(str, optional): Job status to filter by,
        job_type(str, optional): Job type to filter by,
        user_full_name(str, optional): User full name to filter by,
        object_id(str, optional): Object id to filter by,
        sort_by(SortBy, optional): Specifies sorting criteria to
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (string, optional): Custom Error Message for Error Handling
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    params = {
        'nodeName': node_name,  # this needs to be first to work
        'projectId': project_id,
        'status': status,
        'jobType': job_type,
        'userFullName': user_full_name,
        'objectId': object_id,
        'sortBy': sort_by,
        'fields': ','.join(fields) if fields else None,
    }

    params_delete_none = delete_none_values(params, recursion=True)
    params_encoded = urlencode(params_delete_none, True, quote_via=quote)
    return connection.get(endpoint='/api/monitors/jobs', params=params_encoded)


def get_jobs_async(
    future_session: 'FuturesSessionWithRenewal',
    node_name: str,
    project_id: str | None = None,
    status: str | None = None,
    job_type: str | None = None,
    user_full_name: str | None = None,
    object_id: str | None = None,
    sort_by: str | None = None,
    fields: list[str] | None = None,
) -> 'Response':
    """Get list of a jobs asynchronously.

    Args:
        future_session(object): Future Session object to call MicroStrategy REST
            Server asynchronously
        node_name(str): Node name,
        project_id(str, optional): Project id,
        status(str, optional): Job status to filter by,
        job_type(str, optional): Job type to filter by,
        user_full_name(str, optional): User full name to filter by,
        object_id(str, optional): Object id to filter by,
        sort_by(SortBy, optional): Specifies sorting criteria to
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
    Returns:
        FuturesSessionWithRenewal object
    """
    params = {
        'nodeName': node_name,  # this needs to be first to work
        'projectId': project_id,
        'status': status,
        'jobType': job_type,
        'userFullName': user_full_name,
        'objectId': object_id,
        'sortBy': sort_by,
        'fields': ','.join(fields) if fields else None,
    }

    params_delete_none = delete_none_values(params, recursion=True)
    params_encoded = urlencode(params_delete_none, True, quote_via=quote)
    return future_session.get(endpoint='/api/monitors/jobs', params=params_encoded)


@ErrorHandler(err_msg="Error getting jobs list")
def get_jobs_v2(
    connection: 'Connection',
    node_name: str,
    user: list[str] | str | None = None,
    description: str | None = None,
    type: list[str] | str | None = None,
    status: list[str] | str | None = None,
    object_id: list[str] | None = None,
    object_type: list[str] | str | None = None,
    project_id: list[str] | str | None = None,
    project_name: list[str] | str | None = None,
    pu_name: list[str] | str | None = None,
    subscription_type: list[str] | str | None = None,
    subscription_recipient: list[str] | str | None = None,
    memory_usage: str | None = None,
    elapsed_time: str | None = None,
    sort_by: str | None = None,
    fields: list[str] | None = None,
    error_msg: str | None = None,
):
    """Get list of a jobs.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        node_name(str): Node name,
        user(str, optional): Field to filter on job owner's full name (exact
            match),
        description(str, optional): Field to filter on job description (partial
            match),
        type(str, optional): Field to filter on job type (exact match),
        status(str, optional): Job status to filter by,
        object_id(str, optional): Field to filter on object ID of the job
            (exact match)
        object_type(str, optional): Field to filter on object type (exact match)
        project_id(str, optional): Field to filter on project id (exact
            match),
        project_name(str, optional): Field to filter on project name (exact
            match),
        pu_name(str, optional): Field to filter on processing unit name (exact
            match),
        subscription_type(str, optional): Field to filter on subscription type
            (exact match),
        subscription_recipient(str, optional): Field to filter on subscription
            recipient's full name (exact match),
        memory_usage(str, optional): Field to filter on the job elapsed time,
            for example 'gte:100' means filtering jobs with memoryUsage greater
            than or equal to 100 MB. Valid operators are:
            gte - greater than or equal
            lte - less than or equal
        elapsed_time(str, optional): Field to filter on the job elapsed time,
            for example 'gte:100' means filtering jobs with elapsedTime greater
            than or equal to 100 seconds. Valid operators are:
            gte - greater than or equal
            lte - less than or equal
        sort_by(SortBy, optional): Specify sorting criteria, for example
            '+status' means sorting status is ascending order or '-userFullName'
            means sorting userFullName in descending order. Currently, the
            server supports sorting only by single field.
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (string, optional): Custom Error Message for Error Handling
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    params = {
        'nodeName': node_name,  # this needs to be first to work
        'user': user,
        'description': description,
        'type': type,
        'status': status,
        'objectId': object_id,
        'objectType': object_type,
        'projectId': project_id,
        'projectName': project_name,
        'puName': pu_name,
        'subscriptionType': subscription_type,
        'subscriptionRecipient': subscription_recipient,
        'memoryUsage': memory_usage,
        'elapsedTime': elapsed_time,
        'sortBy': sort_by,
        'fields': ','.join(fields) if fields else None,
    }

    params_delete_none = delete_none_values(params, recursion=True)
    params_encoded = urlencode(params_delete_none, True, quote_via=quote)
    return connection.get(endpoint='/api/v2/monitors/jobs', params=params_encoded)


def get_jobs_v2_async(
    future_session: 'FuturesSessionWithRenewal',
    node_name: str,
    user: list[str] | str | None = None,
    description: str | None = None,
    type: list[str] | str | None = None,
    status: list[str] | str | None = None,
    object_id: list[str] | None = None,
    object_type: list[str] | str | None = None,
    project_id: list[str] | str | None = None,
    project_name: list[str] | str | None = None,
    pu_name: list[str] | str | None = None,
    subscription_type: list[str] | str | None = None,
    subscription_recipient: list[str] | str | None = None,
    memory_usage: str | None = None,
    elapsed_time: str | None = None,
    sort_by: str | None = None,
    fields: list[str] | None = None,
) -> 'Response':
    """Get list of a jobs asynchronously.

    Args:
        future_session(object): Future Session object to call MicroStrategy REST
            Server asynchronously
        node_name(str): Node name,
        user(str, optional): Field to filter on job owner's full name (exact
            match),
        description(str, optional): Field to filter on job description (partial
            match),
        type(str, optional): Field to filter on job type (exact match),
        status(str, optional): Job status to filter by,
        object_id(str, optional): Field to filter on object ID of the job
            (exact match)
        object_type(str, optional): Field to filter on object type (exact match)
        project_id(str, optional): Field to filter on project id (exact
            match),
        project_name(str, optional): Field to filter on project name (exact
            match),
        pu_name(str, optional): Field to filter on processing unit name (exact
            match),
        subscription_type(str, optional): Field to filter on subscription type
            (exact match),
        subscription_recipient(str, optional): Field to filter on subscription
            recipient's full name (exact match),
        memory_usage(str, optional): Field to filter on the job elapsed time,
            for example 'gte:100' means filtering jobs with memoryUsage greater
            than or equal to 100 MB. Valid operators are:
            gte - greater than or equal
            lte - less than or equal
        elapsed_time(str, optional): Field to filter on the job elapsed time,
            for example 'gte:100' means filtering jobs with elapsedTime greater
            than or equal to 100 seconds. Valid operators are:
            gte - greater than or equal
            lte - less than or equal
        sort_by(SortBy, optional): Specify sorting criteria, for example
            '+status' means sorting status is ascending order or '-userFullName'
            means sorting userFullName in descending order. Currently, the
            server supports sorting only by single field.
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.

    Returns:
        FuturesSessionWithRenewal object
    """
    params = {
        'nodeName': node_name,  # this needs to be first to work
        'user': user,
        'description': description,
        'type': type,
        'status': status,
        'objectId': object_id,
        'objectType': object_type,
        'projectId': project_id,
        'projectName': project_name,
        'puName': pu_name,
        'subscriptionType': subscription_type,
        'subscriptionRecipient': subscription_recipient,
        'memoryUsage': memory_usage,
        'elapsedTime': elapsed_time,
        'sortBy': sort_by,
        'fields': ','.join(fields) if fields else None,
    }

    params_delete_none = delete_none_values(params, recursion=True)
    params_encoded = urlencode(params_delete_none, True, quote_via=quote)
    return future_session.get(endpoint='/api/v2/monitors/jobs', params=params_encoded)


@ErrorHandler(err_msg="Error killing job {id}")
def cancel_job(
    connection: 'Connection',
    id: str,
    fields: list[str] | None = None,
    error_msg: str | None = None,
):
    """Cancel a job specified by `id`. Use cancel_job_v1 if I-Server version
    is 11.3.2 or cancel_job_v2 otherwise.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id(str): ID of the job
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    if version.parse(connection.iserver_version) == version.parse(
        ISERVER_VERSION_11_3_2
    ):
        return cancel_job_v1(connection, id, fields, error_msg)
    else:
        return cancel_job_v2(connection, id, fields, error_msg)


@ErrorHandler(err_msg="Error killing job {id}")
def cancel_job_v1(
    connection: 'Connection',
    id: str,
    fields: list[str] | None = None,
    error_msg: str | None = None,
):
    """Cancel a job specified by `id`.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id(str): ID of the job
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    params = {'fields': ','.join(fields) if fields else None}

    return connection.delete(endpoint=f'/api/monitors/jobs/{id}', params=params)


@ErrorHandler(err_msg="Error killing job {id}")
def cancel_job_v2(
    connection: 'Connection',
    id: str,
    fields: list[str] | None = None,
    error_msg: str | None = None,
):
    """Cancel a job specified by `id`.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id(str): ID of the job
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg(str, optional): Customized error message.
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    params = {'fields': ','.join(fields) if fields else None}

    return connection.delete(endpoint=f'/api/v2/monitors/jobs/{id}', params=params)


def cancel_jobs(
    connection: 'Connection', ids: list[str], fields: list[str] | None = None
) -> Success | PartialSuccess | MstrException:
    """Cancel jobs specified by `ids`. Use cancel_jobs_v1 if I-Server version
    is 11.3.2 or cancel_jobs_v2 otherwise.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        ids(List[str]): IDs of the jobs
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
    Returns:
        Success: object if all jobs were killed
        PartialSuccess: if not all jobs were killed
        MstrException: otherwise
    """
    if version.parse(connection.iserver_version) == version.parse(
        ISERVER_VERSION_11_3_2
    ):
        response = cancel_jobs_v1(connection, ids, fields)
    else:
        response = cancel_jobs_v2(connection, ids, fields)

    return bulk_operation_response_handler(response, 'jobCancellationStatus')


def cancel_jobs_v1(
    connection: 'Connection', ids: list[str], fields: list[str] | None = None
):
    """Cancel jobs specified by `ids`.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        ids(List[str]): IDs of the jobs
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    params = {'fields': ','.join(fields) if fields else None}

    if ids:
        body = {'jobIds': ids}
        return connection.post(
            endpoint='/api/monitors/cancelJobs', params=params, json=body
        )
    else:
        raise ValueError("No ids have been passed.")


def cancel_jobs_v2(
    connection: 'Connection', ids: list[str], fields: list[str] | None = None
):
    """Cancel jobs specified by `ids`.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        ids(List[str]): IDs of the jobs
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    params = {'fields': ','.join(fields) if fields else None}

    if ids:
        body = {'jobIds': ids}
        return connection.post(
            endpoint='/api/v2/monitors/cancelJobs', params=params, json=body
        )
    else:
        raise ValueError("No ids have been passed.")


@ErrorHandler(err_msg="Error getting caches")
def get_contents_caches(
    connection: 'Connection',
    project_id: str,
    node: str,
    offset: int = 0,
    limit: int = 1000,
    status: str | None = None,
    content_type: str | None = None,
    content_format: str | None = None,
    size: str | None = None,
    owner: str | None = None,
    expiration: str | None = None,
    last_updated: str | None = None,
    hit_count: str | None = None,
    sort_by: str | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
) -> 'Response':
    """Get cache objects

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`,
        project_id(str): Field to filter on project id (exact
            match),
        node(str): Node name,
        offset(int): Starting point within the collection of returned results.
            Used to control paging behavior. Default value = 0,
        limit(int): Maximum number of items returned for a single request.
            Used to control paging behavior. Maximum and default value: 1000,
        status(str, optional): Status of the content cache,
        content_type(str, optional): type of content,
        content_format(str, optional): Format of the content cache, intended for
            dashboard, document and dossier cache,
        size(str, optional): Size of the content cache (in KB),
        owner(str, optional): Owner of the content cache. Exact match on the
            owner's full name,
        expiration(str, optional): Expiration time of the cache,
        last_updated(str, optional): Last update of the cache,
        hit_count(str, optional): Hit count of the cache,
        sort_by(str, optional): Specify sorting criteria,
        fields(str, optional): A whitelist of top-level fields separated by
            commas. Allow the client to selectively retrieve fields in the
            response.
        error_msg (string, optional): Custom Error Message for Error Handling
    Returns:
        HTTP response object. Expected status 200.
    """
    params = {
        'projectId': project_id,
        'clusterNode': node,
        'offset': offset,
        'limit': limit,
        'status': status,
        'type': content_type,
        'format': content_format,
        'size': size,
        'owner': owner,
        'expiration': expiration,
        'lastUpdated': last_updated,
        'hitCount': hit_count,
        'sortBy': sort_by,
        'fields': fields,
    }
    params_delete_none = delete_none_values(params, recursion=True)
    params_encoded = urlencode(params_delete_none, True, quote_via=quote)
    return connection.get(
        endpoint='/api/monitors/caches/contents', params=params_encoded
    )


@ErrorHandler(err_msg='Error updating caches')
def update_contents_caches(
    connection: 'Connection', node: str, body: dict, fields: str | None = None
) -> 'Response':
    """Alter multiple content cache statuses or remove content caches entirely
        in multiple projects at specific node.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`,
        node(str): Node name,
        body(dict): List of contents
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.

    Returns:
        HTTP response object. Expected status 200.
    """
    return connection.patch(
        endpoint='/api/v2/monitors/caches/contents',
        params={'clusterNode': node, 'fields': fields},
        json=body,
    )


@ErrorHandler(err_msg='Error deleting caches')
def delete_caches(connection: 'Connection', project_id: str, cache_type: str):
    """Delete element or object caches for a specific project.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`,
        project_id(str): Project id,
        cache_type(str): Cache type
            Available values:
            - 'element'
            - 'object'

    Returns:
        HTTP response object. Expected status 204.
    """

    return connection.delete(
        endpoint=f'/api/monitors/projects/{project_id}/caches/{cache_type}',
    )



"""This is the demo script to show how administrator can manage project
languages.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""
from mstrio.connection import get_connection
from mstrio.datasources import list_datasource_connections
from mstrio.server import Project
from mstrio.server.project_languages import (
    CurrentMode,
    DataLocalizationLanguage,
    SimpleLanguage,
)

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Get project with a given name
project = Project(connection=conn, name=PROJECT_NAME)

# Get a default data language of a project
data_default = project.data_language_settings.default_language

# Get a current mode for data internationalization of a project
data_mode = project.data_language_settings.current_mode

# Get an available data languages of a project
data_languages = project.data_language_settings.languages

# Get a default display language of a project
metadata_default = project.metadata_language_settings.default

# Get an available metadata languages of a project
metadata_languages = project.metadata_language_settings.languages

# Get a list of datasource connections
# might be useful to get connection_id for data language
datasource_connections = list_datasource_connections(conn)

# Data internalization

# Define variables which can be later used in a script
DATA_LANGUAGE_ID_1 = $data_language_id_1
DATA_LANGUAGE_NAME_1 = $data_language_name_1
DATA_LANGUAGE_COLUMN_1 = $data_language_column_1
DATA_LANGUAGE_TABLE_1 = $data_language_table_1
DATA_LANGUAGE_COLUMN_1_ALTERED = $data_language_column_1_altered
# Could be one of the following: 'column', 'table', 'connectionId'
DATA_ALTER_PATH = $alter_path

DATA_LANGUAGE_ID_2 = $data_language_id_2
DATA_LANGUAGE_NAME_2 = $data_language_name_2
DATA_LANGUAGE_CONNECTION_ID_2 = $data_language_connection_id_2

# Add data languages to a project
data_languages_to_add = [
    DataLocalizationLanguage(
        id=DATA_LANGUAGE_ID_1,
        name=DATA_LANGUAGE_NAME_1,
        column=DATA_LANGUAGE_COLUMN_1,
        table=DATA_LANGUAGE_TABLE_1,
    ),
    DataLocalizationLanguage(
        id=DATA_LANGUAGE_ID_2,
        name=DATA_LANGUAGE_NAME_2,
        connection_id=DATA_LANGUAGE_CONNECTION_ID_2,
    )
]

project.data_language_settings.add_language(languages=data_languages_to_add)

# Alter data language of a project
data_language_to_alter = [
    DataLocalizationLanguage(
        id=DATA_LANGUAGE_ID_1, column=DATA_LANGUAGE_COLUMN_1_ALTERED
    )
]

project.data_language_settings.alter_language(
    languages=data_language_to_alter, path=DATA_ALTER_PATH
)

# Remove data languages from a project
data_languages_to_remove = [
    DataLocalizationLanguage(id=DATA_LANGUAGE_ID_1),
    DataLocalizationLanguage(id=DATA_LANGUAGE_ID_2)
]

project.data_language_settings.remove_language(languages=data_languages_to_remove)

# Set default data language of a project
project.data_language_settings.alter_default_language(language=DATA_LANGUAGE_ID_1)

# Set current mode of a data project
project.data_language_settings.alter_current_mode(mode=CurrentMode.NONE)

# Metadata internalization

# Define variables which can be later used in a script
META_LANGUAGE_ID_1 = $meta_language_id_1
META_LANGUAGE_NAME_1 = $meta_language_name_1

META_LANGUAGE_ID_2 = $meta_language_id_2
META_LANGUAGE_NAME_2 = $meta_language_name_2

# Add metadata languages to a project
metadata_languages_to_add = [
    SimpleLanguage(id=META_LANGUAGE_ID_1, name=META_LANGUAGE_NAME_1),
    SimpleLanguage(id=META_LANGUAGE_ID_2, name=META_LANGUAGE_NAME_2),
]

project.metadata_language_settings.add_language(languages=metadata_languages_to_add)

# Remove metadata languages from a project
metadata_languages_to_remove = [
    SimpleLanguage(id=META_LANGUAGE_ID_1),
    SimpleLanguage(id=META_LANGUAGE_ID_2),
]

project.metadata_language_settings.remove_language(languages=metadata_languages_to_remove)



# This example is for a dashboard user to modify data in a PostgreSQL DB table using 
# a modern grid configured with transaction functions. The workflow will rely on this 
# script to make actual modifications in the table. The table named USERS has 4 
# columns USER_ID(Int), USER_LOGIN(Varchar), HIRED_DATE(Date) and EMAIL(VarChar) 
# with USER_ID as the primary. Data changes for HIRED_DATE (Date) and EMAIL(VarChar) 
# are optional.

# In this template we use the library 'psycopg2' (installed in runtime as
# 'psycopg2-binary') to connect with the PostgreSQL database, submit atomic requests 
# and execute operations. For using other types of database, make sure to use
# their methods correctly. 

# Python libraries to handle SQL connections are not installed on the environment by
# default so it is necessary to choose runtime which have such library installed in the
# settings of the script. Additionally it is necessary to specify correct `Network
# Access` to allow script to communicate with some other IP then IP of the environment.
import psycopg2
import datetime


try:
    # create connection to database. If you want to use other way then providing
    # user and password please check documentation of this method 
    connection = None
    connection = psycopg2.connect(
        host=$host,
        port=$port,
        database=$database,
        user=$user_name,
        password=$user_password
    )

    # Prepare parameterized query for select the existing row to delete
    sql_delete_query = """DELETE FROM USERS WHERE USER_ID = %s"""

    # create cursor to execute operations on database
    cursor = connection.cursor()
    
    # deal with the every row data via prepared statements
    # we assume that the logic of passing data from dashboard works correctly and each
    # transactional column is either the array of length which is equal to number
    # of rows passed from dashboard
    for index in range(len($user_id)):
        # case with DELETE query is the easiest one because we are deleting
        # just based on `USER_ID`
        cursor.execute(sql_delete_query, tuple([int($user_id[index])]))

    # commit all changes when there weren't any errors
    connection.commit()

# error handling
except psycopg2.Error as e:
    # in case of any error we have to rollback all of the changes which
    # has been made on database
    if connection:
        connection.rollback()
    print("Error:", e)

finally:
    # close the database connection
    if connection:
        connection.close()
        


"""This is the demo script to show how to manage events.
This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.distribution_services import Event, list_events
from mstrio.connection import get_connection

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# List all events
all_events = list_events(conn)

# Define variables which can be later used in a script
EVENT_ID = $event_id  # id for Event object
EVENT_NAME = $event_name  # name for Event object

# Get event by id
sample_event = Event(conn, id=EVENT_ID)

# Get event by name
sample_event = Event(conn, name=EVENT_NAME)

# list the Event's properties
sample_event.list_properties()

# Define a variable which can be later used in a script
EVENT_DESCRIPTION = $event_description

# Create event with description
db_load_event = Event.create(conn, name=EVENT_NAME, description=EVENT_DESCRIPTION)

# Trigger the event
db_load_event.trigger()

# Rename the event via alter
db_load_event.alter(name=EVENT_NAME, description=EVENT_DESCRIPTION)

# Delete the event
# Please note that when argument `force` is set to `False` (default value),
# deletion must be confirmed by selecting the appropriate prompt value.
# Do also note that the Event object stored in the `db_load_event` variable is
# still valid after the rename above, as it internally references the event
# by ID, not name.
db_load_event.delete(force=True)



import logging
from collections.abc import Callable
from enum import Enum, auto
from pathlib import Path

from mstrio import config
from mstrio.connection import Connection
from mstrio.helpers import VersionException
from mstrio.object_management.migration.package import (
    Package,
    PackageConfig,
    PackageImport,
)
from mstrio.utils.helper import Dictable, exception_handler
from mstrio.utils.progress_bar_mixin import ProgressBarMixin
from mstrio.utils.wip import WipLevels, module_wip

module_wip(globals(), level=WipLevels.PREVIEW)

logger = logging.getLogger(__name__)


class MigrationStatus(Enum):
    NOT_STARTED = auto()
    STARTED = auto()
    CREATING_PACKAGE = auto()
    CREATING_PACKAGE_FAILED = auto()
    CREATING_PACKAGE_COMPLETED = auto()
    MIGRATION_IN_PROGRESS = auto()
    MIGRATION_FAILED = auto()
    MIGRATION_COMPLETED = auto()
    UNDO_STARTED = auto()
    UNDO_FAILED = auto()
    UNDO_COMPLETED = auto()


class Migration(Dictable, ProgressBarMixin):
    """A class encapsulating migration process from env A to env B.

    Raises VersionException if either of environments are running IServer
    version lower than 11.3.0200

    Raises AttributeError if parent directory of specified `save_path` is
    non-existent or points to another kind of file.

    Attributes:
        configuration(PackageConfig): a configuration of the migration
        save_path(str): a full path used when saving import package to a
        file. It is also used for saving undo package with the addition of
        `_undo` suffix. By default, uses ~cwd/migration.mmp and
        ~cwd/migration_undo.mmp.
        source_connection(Connection): A MicroStrategy connection object
          to the source env,
        target_connection(Connection): A MicroStrategy connection object
          to the target env,
        name(str): Name of the migration. Used for identification purposes for
            the convenience of the user,
        package(Package): a package that is migrated,
        package_import(PackageImport): a package import process object,
        package_binary(Binary): actual data that is migrated
        status(MigrationStatus): current status of migration,
        create_undo(bool): should a undo package be made
        undo_package(Binary): undo package binary for this import process.
          if `create_undo` is False, it is equal to None.

    Additional args:
        custom_package_path(str): a full path to an already existing .mpp file.
            This argument overrides `save_path`. Creating a `Migration` object
            with only `source_connection`, `target_connection` and
            `custom_package_path` allows for an easier export to the target by
            running `migrate_package()` on that `Migration` instance (instead of
            performing full migration with an import from the source first.
            See demos and examples for details.)."""

    def __init__(
        self,
        save_path: str | None = None,
        source_connection: Connection | None = None,
        configuration: PackageConfig | None = None,
        target_connection: Connection | None = None,
        custom_package_path: str | None = None,
        name: str | None = None,
        create_undo: bool = True,
    ):
        save_path = save_path or (Path.cwd() / 'migration.mmp')
        if not self._validate_envs_version(source_connection, target_connection):
            exception_handler(
                msg=(
                    "Environments must run IServer version 11.3.0200 or newer. "
                    "Please update your environments to use this feature."
                ),
                exception_type=VersionException,
            )
        self.name = name
        self.source_connection = source_connection
        self.target_connection = target_connection
        self.configuration = configuration
        self.create_undo = create_undo
        self._status = MigrationStatus.NOT_STARTED

        self._check_file_path(save_path, "save_path")
        self.save_path = save_path

        self._package = None
        self._package_import = None
        self._undo_binary = None

        if custom_package_path and self._check_file_path(
            custom_package_path, "custom_package_path"
        ):
            self.save_path = custom_package_path
            filename, file_extension = self._decompose_file_path(custom_package_path)
            with open(f"{filename}{file_extension}", "rb") as f:
                self._package_binary = f.read()
        else:
            self._package_binary = None

    @staticmethod
    def _validate_envs_version(
        source_connection: Connection, target_connection: Connection
    ) -> bool:
        return (
            source_connection is None
            or source_connection._iserver_version >= '11.3.0200'
        ) and (
            target_connection is None
            or target_connection._iserver_version >= '11.3.0200'
        )

    @staticmethod
    def _decompose_file_path(file_path: str) -> tuple:
        file_path = Path(file_path)
        filename, file_extension = file_path.parent / file_path.stem, file_path.suffix
        if file_extension == '':
            return filename, '.mmp'
        return filename, file_extension

    @staticmethod
    def _check_file_path(file_path: str, var_name: str):
        file_path = Path(file_path)

        if file_path.parent.is_dir() and not file_path.is_dir():
            return True
        else:
            exception_handler(
                msg=f"Invalid save path. Parent directory specified in `{var_name}` is"
                f" incorrect.",
                exception_type=AttributeError,
            )

    def perform_full_migration(self) -> bool:
        """Perform 'create_package()' and 'migrate_package()' using
        configuration provided when creating `Migration` object.
        """
        if (
            not isinstance(self.source_connection, Connection)
            or self.source_connection is None
        ):
            exception_handler(
                msg=(
                    "Migration object missing `source_connection`. "
                    "`perform_full_migration()` unavailable."
                ),
                exception_type=AttributeError,
            )
        if (
            not isinstance(self.target_connection, Connection)
            or self.target_connection is None
        ):
            exception_handler(
                msg=(
                    "Migration object missing `target_connection`. "
                    "`perform_full_migration()` unavailable."
                ),
                exception_type=AttributeError,
            )

        self._display_progress_bar(
            desc='Migration status: ',
            unit='Migration',
            total=4,
            bar_format='{desc} |{bar:50}| {percentage:3.0f}%',
        )

        if not self.create_package():
            return False

        if not self.migrate_package():
            return False

        self._close_progress_bar()
        return True

    def create_package(self) -> bool:
        """Performs import of the object described in the migration
        configuration from the source environment and save it in a file
        at location specified in `save_path` parameter.

        Raises AttributeError if `source_connection` is not specified.

        Raises FileExistsError if a package or undo package with the same name
        already exist at`save_path` location."""
        if (
            not isinstance(self.source_connection, Connection)
            or self.source_connection is None
        ):
            exception_handler(
                msg=(
                    "Migration object does not have `source_connection`. Import "
                    "unavailable."
                ),
                exception_type=AttributeError,
            )

        filename, file_extension = self._decompose_file_path(self.save_path)
        """File existence checked here instead of using "xb" mode in context
        manager to avoid creation of package if it should not be saved"""
        if (
            Path(f"{filename}{file_extension}").exists()
            or Path(f"{filename}_undo{file_extension}").exists()
        ):
            exception_handler(
                msg=(
                    "Migration file / undo package with this name already exists."
                    "Please use other location / filename in `save_path` parameter."
                ),
                exception_type=FileExistsError,
            )

        self.__private_status = MigrationStatus.CREATING_PACKAGE
        self._create_package_holder(self.source_connection)
        if not self._update_package_holder():
            self.__private_status = MigrationStatus.CREATING_PACKAGE_FAILED
            return False

        self._download_package_binary()
        self._delete_package_holder()
        self._save_package_binary_locally(filename, file_extension, self.package_binary)

        self.__private_status = MigrationStatus.CREATING_PACKAGE_COMPLETED
        return True

    def migrate_package(
        self, custom_package_path: str | None = None, is_undo=False
    ) -> bool:
        """Performs migration of already created package to the target
        environment. Import package will be loaded from `custom_package_path`.
        If `custom_package_path` not provided, the object previously acquired
        with the `create_package()` will be used.
        If `create_undo` parameter is set to True, package needed for undo
        process will be downloaded.

        Raises AttributeError if `target_connection` is not specified.
        """

        if (
            not isinstance(self.target_connection, Connection)
            or self.target_connection is None
        ):
            exception_handler(
                msg=(
                    "Migration object does not have `target_connection`. "
                    "Export unavailable."
                ),
                exception_type=AttributeError,
            )

        if custom_package_path and self._check_file_path(
            custom_package_path, "custom_package_path"
        ):
            self.save_path = custom_package_path
            filename, file_extension = self._decompose_file_path(custom_package_path)
            with open(f"{filename}{file_extension}", "rb") as f:
                return self._migrate_package(f.read(), is_undo=is_undo)
        return self._migrate_package(
            self._package_binary,
            custom_package_path=custom_package_path,
            is_undo=is_undo,
        )

    def _migrate_package(
        self,
        binary: bytes,
        is_undo: bool = False,
        custom_package_path: str | None = None,
    ) -> bool:
        if binary is None:
            exception_handler(
                msg=(
                    "Import package is None. Run `create_package()` first, "
                    "or specify `custom_package_path`."
                ),
                exception_type=AttributeError,
            )

        self.__private_status = MigrationStatus.MIGRATION_IN_PROGRESS
        self._create_package_holder(self.target_connection)
        self._upload_package_binary(binary)
        if not self._create_import(self.target_connection):
            self.__private_status = MigrationStatus.MIGRATION_FAILED
            return False

        if not is_undo and self.create_undo:
            self._download_undo_binary()
            file_path = custom_package_path if custom_package_path else self.save_path
            filename, file_extension = self._decompose_file_path(file_path)
            self._save_package_binary_locally(
                filename=f"{filename}_undo",
                file_extension=file_extension,
                _bytes=self.undo_binary,
            )

        self._delete_package_holder()
        self._delete_import()

        self.__private_status = MigrationStatus.MIGRATION_COMPLETED
        return True

    def undo_migration(self):
        """Revert the migration using the package downloaded during
        `migrate_package()` or `perform_full_migration()`.

        Raises AttributeError if `undo_binary` is None.
        """
        self.__private_status = MigrationStatus.UNDO_STARTED
        if self.undo_binary:
            self._migrate_package(self._undo_binary, True)
        else:
            self.__private_status = MigrationStatus.UNDO_FAILED
            msg = (
                '`undo_binary` is None. Perform migration with `create_undo`'
                ' parameter set to True'
            )
            exception_handler(msg, exception_type=AttributeError)
        self.__private_status = MigrationStatus.UNDO_COMPLETED

    def _create_package_holder(self, conn: Connection):
        self._package = Package.create(
            conn, progress_bar=self._progress_bar is not None
        )

    def _update_package_holder(self):
        return self._package.update_config(self.configuration)

    def _download_package_binary(self):
        self._package_binary = self._package.download_package_binary(
            progress_bar=self._progress_bar is not None
        )

    def _upload_package_binary(self, binary: bytes):
        self._package.upload_package_binary(
            binary, progress_bar=self._progress_bar is not None
        )

    def _delete_package_holder(self):
        self._package.delete(force=True)
        self._package = None

    def _create_import(self, conn: Connection) -> bool:
        self._package_import = PackageImport.create(
            conn,
            self.package.id,
            self.create_undo,
            progress_bar=self._progress_bar is not None,
        )
        if self._package_import is False:
            return False
        return True

    def _delete_import(self):
        self._package_import.delete(force=True)
        self._package_import = None

    def _download_undo_binary(self):
        self._undo_binary = self._package_import.download_undo_binary(
            progress_bar=self._progress_bar is not None
        )

    def _save_package_binary_locally(
        self, filename: str, file_extension: str, _bytes: bytes
    ):
        with open(f"{filename}{file_extension}", "wb") as f:
            f.write(_bytes)
        logger.info(
            f'Package / package undo binary created at:{filename}{file_extension}'
        )

    @property
    def package(self):
        return self._package

    @property
    def package_binary(self):
        return self._package_binary

    @property
    def package_import(self):
        return self._package_import

    @property
    def undo_binary(self):
        return self._undo_binary

    @property
    def status(self):
        return self._status

    @property
    def __private_status(self):
        return self._status

    @__private_status.setter
    def __private_status(self, var: MigrationStatus):
        self._update_progress_bar_if_needed(
            new_description=f'[{self.name}] Status: {var.name}'
        )
        self._status = var


def bulk_full_migration(migrations: Migration | list[Migration], verbose: bool = False):
    """Run `perform_full_migration()` for each of the migrations provided.
    Args:
        migrations: migrations to be executed
        verbose: if True, information about each step will be printed
    """
    __run_func_for_all_migrations(Migration.perform_full_migration, migrations, verbose)


def bulk_migrate_package(
    migrations: Migration | list[Migration], verbose: bool = False
):
    """Run `migrate_package()` for each of the migrations provided.
    Args:
        migrations: migrations to be executed
        verbose: if True, information about each step will be printed
    """
    __run_func_for_all_migrations(Migration.migrate_package, migrations, verbose)


def __run_func_for_all_migrations(
    func: Callable,
    migrations: Migration | list[Migration],
    verbose: bool = False,
    *args,
    **kwargs,
):
    if type(migrations) is Migration:
        migrations = [migrations]
    original_verbose = config.verbose
    config.verbose = verbose
    for migration in migrations:
        func(migration, *args, **kwargs)
    config.verbose = original_verbose
